import math

from torch.utils.data.sampler import BatchSampler
from torch.utils.data.sampler import Sampler
from torch.utils.data.sampler import SubsetRandomSampler


def identity(x):
    return x


class SortedSampler(Sampler):
    """
    Samples elements sequentially, always in the same order.

    Args:
        data (`obj`: `Iterable`):
            Iterable data.
        sort_key (`obj`: `Callable`):
            Specifies a function of one argument that is used to
            extract a numerical comparison key from each list element.

        Example:
            >>> list(SortedSampler(range(10), sort_key=lambda i: -i))
                [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]

    """

    def __init__(self, data, sort_key=identity):
        super().__init__(data)
        self.data = data
        self.sort_key = sort_key
        zip_ = [(i, self.sort_key(row)) for i, row in enumerate(self.data)]
        zip_ = sorted(zip_, key=lambda r: r[1])
        self.sorted_indexes = [item[0] for item in zip_]

    def __iter__(self):
        return iter(self.sorted_indexes)

    def __len__(self):
        return len(self.data)


class BucketBatchSampler(BatchSampler):
    """
    `BucketBatchSampler` toggles between `sampler` batches and sorted batches.
    Typically, the `sampler` will be a `RandomSampler` allowing the user to toggle between
    random batches and sorted batches. A larger `bucket_size_multiplier` is more sorted and vice
    versa.
    Background:
        ``BucketBatchSampler`` is similar to a ``BucketIterator`` found in popular libraries like
        ``AllenNLP`` and ``torchtext``. A ``BucketIterator`` pools together examples with a similar
        size length to reduce the padding required for each batch while maintaining some noise
        through bucketing.
        **AllenNLP Implementation:**
        https://github.com/allenai/allennlp/blob/master/allennlp/data/iterators/bucket_iterator.py
        **torchtext Implementation:**
        https://github.com/pytorch/text/blob/master/torchtext/data/iterator.py#L225

    Args:
        sampler (`obj`: `torch.data.utils.sampler.Sampler):
        batch_size (`int`):
            Size of mini-batch.
        drop_last (`bool`, optional, defaults to `False`):
            If `True` the sampler will drop the last batch if its size would be less than `batch_size`.
        sort_key (`obj`: `Callable`, optional, defaults to `identity`):
            Callable to specify a comparison key for sorting.
        bucket_size_multiplier (`int`, optional, defaults to `100`):
            Buckets are of size `batch_size * bucket_size_multiplier`.
    Example:
        >>> from torchnlp.random import set_seed
        >>> set_seed(123)
        >>>
        >>> from torch.utils.data.sampler import SequentialSampler
        >>> sampler = SequentialSampler(list(range(10)))
        >>> list(BucketBatchSampler(sampler, batch_size=3, drop_last=False))
        [[6, 7, 8], [0, 1, 2], [3, 4, 5], [9]]
        >>> list(BucketBatchSampler(sampler, batch_size=3, drop_last=True))
        [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

    """

    def __init__(
        self,
        sampler,
        batch_size,
        drop_last: bool = False,
        sort_key=identity,
        bucket_size_multiplier=100,
    ):
        super().__init__(sampler, batch_size, drop_last)
        self.sort_key = sort_key
        _bucket_size = batch_size * bucket_size_multiplier
        if hasattr(sampler, "__len__"):
            _bucket_size = min(_bucket_size, len(sampler))
        self.bucket_sampler = BatchSampler(sampler, _bucket_size, False)

    def __iter__(self):
        for bucket in self.bucket_sampler:
            sorted_sampler = SortedSampler(bucket, self.sort_key)
            for batch in SubsetRandomSampler(
                list(BatchSampler(sorted_sampler, self.batch_size, self.drop_last))
            ):
                yield [bucket[i] for i in batch]

    def __len__(self):
        if self.drop_last:
            return len(self.sampler) // self.batch_size
        else:
            return math.ceil(len(self.sampler) / self.batch_size)



import torch
import math
import numpy as np

from torch.nn import init
from torch import nn
from transformers import AutoModel, AutoConfig

from maverick.common.util import *
from maverick.common.constants import *


class Maverick_mes(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        # document transformer encoder
        self.encoder_hf_model_name = kwargs["huggingface_model_name"]
        self.encoder = AutoModel.from_pretrained(self.encoder_hf_model_name)
        self.encoder_config = AutoConfig.from_pretrained(self.encoder_hf_model_name)
        self.encoder.resize_token_embeddings(self.encoder.embeddings.word_embeddings.num_embeddings + 3)

        # freeze
        if kwargs["freeze_encoder"]:
            for param in self.encoder.parameters():
                param.requires_grad = False

        # span representation, now is concat_start_end
        self.span_representation = kwargs["span_representation"]
        # type of representation layer in 'Linear, FC, LSTM-left, LSTM-right, Conv1d'
        self.representation_layer_type = "FC"  # fullyconnected
        # span hidden dimension
        self.token_hidden_size = self.encoder_config.hidden_size

        # if span representation method is to concatenate start and end, a mention hidden size will be 2*token_hidden_size
        if self.span_representation == "concat_start_end":
            self.mention_hidden_size = self.token_hidden_size * 2

        self.num_cats = len(CATEGORIES) + 1  # +1 for ALL
        self.all_cats_size = self.token_hidden_size * self.num_cats

        self.coref_start_all_mlps = RepresentationLayer(
            type=self.representation_layer_type,  # fullyconnected
            input_dim=self.token_hidden_size,
            output_dim=self.all_cats_size,
            hidden_dim=self.mention_hidden_size,
        )

        self.coref_end_all_mlps = RepresentationLayer(
            type=self.representation_layer_type,  # fullyconnected
            input_dim=self.token_hidden_size,
            output_dim=self.all_cats_size,
            hidden_dim=self.mention_hidden_size,
        )

        self.antecedent_s2s_all_weights = nn.Parameter(
            torch.empty((self.num_cats, self.token_hidden_size, self.token_hidden_size))
        )
        self.antecedent_e2e_all_weights = nn.Parameter(
            torch.empty((self.num_cats, self.token_hidden_size, self.token_hidden_size))
        )
        self.antecedent_s2e_all_weights = nn.Parameter(
            torch.empty((self.num_cats, self.token_hidden_size, self.token_hidden_size))
        )
        self.antecedent_e2s_all_weights = nn.Parameter(
            torch.empty((self.num_cats, self.token_hidden_size, self.token_hidden_size))
        )

        self.antecedent_s2s_all_biases = nn.Parameter(torch.empty((self.num_cats, self.token_hidden_size)))
        self.antecedent_e2e_all_biases = nn.Parameter(torch.empty((self.num_cats, self.token_hidden_size)))
        self.antecedent_s2e_all_biases = nn.Parameter(torch.empty((self.num_cats, self.token_hidden_size)))
        self.antecedent_e2s_all_biases = nn.Parameter(torch.empty((self.num_cats, self.token_hidden_size)))

        # mention extraction layers
        # representation of start token
        self.start_token_representation = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        # representation of end token
        self.end_token_representation = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        # models probability to be the start of a mention
        self.start_token_classifier = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=1,
            hidden_dim=self.token_hidden_size,
        )

        # model mention probability from start and end representations
        self.start_end_classifier = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.mention_hidden_size,
            output_dim=1,
            hidden_dim=self.token_hidden_size,
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        W = [
            self.antecedent_s2s_all_weights,
            self.antecedent_e2e_all_weights,
            self.antecedent_s2e_all_weights,
            self.antecedent_e2s_all_weights,
        ]

        B = [
            self.antecedent_s2s_all_biases,
            self.antecedent_e2e_all_biases,
            self.antecedent_s2e_all_biases,
            self.antecedent_e2s_all_biases,
        ]

        for w, b in zip(W, B):
            init.kaiming_uniform_(w, a=math.sqrt(5))
            fan_in, _ = init._calculate_fan_in_and_fan_out(w)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(b, -bound, bound)

    # takes last_hidden_states, eos_mask, ground truth and stage
    def eos_mention_extraction(self, lhs, eos_mask, gold_mentions, gold_starts, stage):
        start_idxs = []
        mention_idxs = []
        start_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)
        mention_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)

        for bidx in range(0, lhs.shape[0]):
            lhs_batch = lhs[bidx]  # SEQ_LEN X HIDD_DIM
            eos_mask_batch = eos_mask[bidx]  # SEQ_LEN X SEQ_LEN

            # compute start logits
            start_logits_batch = self.start_token_classifier(lhs_batch).squeeze(-1)  # SEQ_LEN

            if gold_starts != None:
                loss = torch.nn.functional.binary_cross_entropy_with_logits(start_logits_batch, gold_starts[bidx])

                # accumulate loss
                start_loss = start_loss + loss

            # compute start positions
            start_idxs_batch = ((torch.sigmoid(start_logits_batch) > 0.5)).nonzero(as_tuple=False).squeeze(-1)

            start_idxs.append(start_idxs_batch.detach().clone())
            # in training, use gold starts to learn to extract mentions, inference use predicted ones
            if stage == "train":
                start_idxs_batch = (
                    ((torch.sigmoid(gold_starts[bidx]) > 0.5)).nonzero(as_tuple=False).squeeze(-1)
                )  # NUM_GOLD_STARTS

            # contains all possible start end indices pairs, i.e. for all starts, all possible ends looking at EOS index
            possibles_start_end_idxs = (eos_mask_batch[start_idxs_batch] == 1).nonzero(as_tuple=False)  # STARTS x 2

            # this is to have reference respect to original positions
            possibles_start_end_idxs[:, 0] = start_idxs_batch[possibles_start_end_idxs[:, 0]]

            possible_start_idxs = possibles_start_end_idxs[:, 0]
            possible_end_idxs = possibles_start_end_idxs[:, 1]

            # extract start and end hidden states
            starts_hidden_states = lhs_batch[possible_end_idxs]  # start
            ends_hidden_states = lhs_batch[possible_start_idxs]  # end

            # concatenation of start to end representations created using a representation layer
            s2e_representations = torch.cat(
                (
                    self.start_token_representation(starts_hidden_states),
                    self.end_token_representation(ends_hidden_states),
                ),
                dim=-1,
            )

            # classification of mentions
            s2e_logits = self.start_end_classifier(s2e_representations).squeeze(-1)

            # mention_start_idxs and mention_end_idxs
            mention_idxs.append(possibles_start_end_idxs[torch.sigmoid(s2e_logits) > 0.5].detach().clone())

            if s2e_logits.shape[0] != 0 and stage != "test":
                if gold_mentions != None:
                    mention_loss_batch = torch.nn.functional.binary_cross_entropy_with_logits(
                        s2e_logits,
                        gold_mentions[bidx][possible_start_idxs, possible_end_idxs],
                    )
                    mention_loss = mention_loss + mention_loss_batch

        return (start_idxs, mention_idxs, start_loss, mention_loss)

    def _get_cluster_labels_after_pruning(self, span_starts, span_ends, all_clusters):
        """
        :param span_starts: [batch_size, max_k]
        :param span_ends: [batch_size, max_k]
        :param all_clusters: [batch_size, max_cluster_size, max_clusters_num, 2]
        :return: [batch_size, max_k, max_k + 1] - [b, i, j] == 1 if i is antecedent of j
        """
        span_starts = span_starts.unsqueeze(0)
        span_ends = span_ends.unsqueeze(0)
        batch_size, max_k = span_starts.size()
        new_cluster_labels = torch.zeros((batch_size, max_k, max_k), device="cpu")
        all_clusters_cpu = all_clusters.cpu().numpy()
        for b, (starts, ends, gold_clusters) in enumerate(
            zip(span_starts.cpu().tolist(), span_ends.cpu().tolist(), all_clusters_cpu)
        ):
            gold_clusters = self.extract_clusters(gold_clusters)
            mention_to_gold_clusters = extract_mentions_to_clusters(gold_clusters)
            gold_mentions = set(mention_to_gold_clusters.keys())
            for i, (start, end) in enumerate(zip(starts, ends)):
                if (start, end) not in gold_mentions:
                    continue
                for j, (a_start, a_end) in enumerate(list(zip(starts, ends))[:i]):
                    if (a_start, a_end) in mention_to_gold_clusters[(start, end)]:
                        new_cluster_labels[b, i, j] = 1
        new_cluster_labels = new_cluster_labels.to(self.encoder.device)
        return new_cluster_labels

    def _get_all_labels(self, clusters_labels, categories_masks):
        batch_size, max_k, _ = clusters_labels.size()

        categories_labels = clusters_labels.unsqueeze(1).repeat(1, self.num_cats, 1, 1) * categories_masks
        all_labels = categories_labels
        # null cluster
        # zeros = torch.zeros((batch_size, self.num_cats, max_k, 1), device=self.encoder.device)
        # all_labels = torch.cat((all_labels, zeros), dim=-1)  # [batch_size, num_cats + 1, max_k, max_k + 1]
        # no_antecedents = 1 - torch.sum(all_labels, dim=-1).bool().float()
        # all_labels[:, :, :, -1] = no_antecedents

        return all_labels

    def mes_span_clustering(
        self, mention_start_reps, mention_end_reps, mention_start_idxs, mention_end_idxs, gold, stage, mask, add, sing
    ):
        coref_logits = self._calc_coref_logits(mention_start_reps, mention_end_reps)
        coref_logits = coref_logits[0] * mask[0]
        coreference_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)

        coref_logits = torch.stack([matrix.tril().fill_diagonal_(0) for matrix in coref_logits]).unsqueeze(0)
        if stage == "train":
            labels = self._get_cluster_labels_after_pruning(mention_start_idxs, mention_end_idxs, gold)
            all_labels = self._get_all_labels(labels, mask)
            coreference_loss = torch.nn.functional.binary_cross_entropy_with_logits(coref_logits, all_labels)

        coref_logits = coref_logits.sum(dim=1)
        doc, m2a, singletons = self.create_mention_to_antecedent_singletons(mention_start_idxs, mention_end_idxs, coref_logits)
        if not sing:
            singletons = []
        coreferences = self.create_clusters(m2a, singletons, add)
        return coreference_loss, coreferences

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_cats, self.token_hidden_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)  # bnkf/bnlg

    def _calc_coref_logits(self, top_k_start_coref_reps, top_k_end_coref_reps):
        all_starts = self.transpose_for_scores(self.coref_start_all_mlps(top_k_start_coref_reps))
        all_ends = self.transpose_for_scores(self.coref_end_all_mlps(top_k_end_coref_reps))

        logits = (
            torch.einsum("bnkf, nfg, bnlg -> bnkl", all_starts, self.antecedent_s2s_all_weights, all_starts)
            + torch.einsum("bnkf, nfg, bnlg -> bnkl", all_ends, self.antecedent_e2e_all_weights, all_ends)
            + torch.einsum("bnkf, nfg, bnlg -> bnkl", all_starts, self.antecedent_s2e_all_weights, all_ends)
            + torch.einsum("bnkf, nfg, bnlg -> bnkl", all_ends, self.antecedent_e2s_all_weights, all_starts)
        )

        biases = (
            torch.einsum("bnkf, nf -> bnk", all_starts, self.antecedent_s2s_all_biases).unsqueeze(-2)
            + torch.einsum("bnkf, nf -> bnk", all_ends, self.antecedent_e2e_all_biases).unsqueeze(-2)
            + torch.einsum("bnkf, nf -> bnk", all_ends, self.antecedent_s2e_all_biases).unsqueeze(-2)
            + torch.einsum("bnkf, nf -> bnk", all_starts, self.antecedent_e2s_all_biases).unsqueeze(-2)
        )

        return logits + biases

    def _get_categories_labels(self, tokens, subtoken_map, new_token_map, span_starts, span_ends):
        max_k = span_starts.shape[0]

        doc_spans = []
        for start, end in zip(span_starts, span_ends):
            token_indices = [new_token_map[0][idx] for idx in set(subtoken_map[0][start : end + 1]) - {None}]
            span = {tokens[0][idx].lower() for idx in token_indices if idx is not None}
            pronoun_id = get_pronoun_id(span)
            doc_spans.append((span - STOPWORDS, pronoun_id))

        categories_labels = np.zeros((max_k, max_k)) - 1
        for i in range(max_k):
            for j in list(range(max_k))[:i]:
                categories_labels[i, j] = get_category_id(doc_spans[i], doc_spans[j])

        categories_labels = torch.tensor(categories_labels, device=self.encoder.device).unsqueeze(0)
        categories_masks = [categories_labels == cat_id for cat_id in range(self.num_cats - 1)] + [categories_labels != -1]
        categories_masks = torch.stack(categories_masks, dim=1).int()
        return categories_labels, categories_masks

    def create_mention_to_antecedent_singletons(self, span_starts, span_ends, coref_logits):
        span_starts = span_starts.unsqueeze(0)
        span_ends = span_ends.unsqueeze(0)
        bs, n_spans, _ = coref_logits.shape
        # long distance regularization
        # a = torch.sigmoid(coref_logits)
        # m1 = torch.arange(coref_logits.shape[1], device=a.device).unsqueeze(0).repeat(coref_logits.shape[1], 1)
        # m2 = m1.transpose(0, 1)
        # m = (torch.ones_like(coref_logits[0], device=a.device) + (m2 - m1 - 1) / 1000).tril().fill_diagonal_(0).unsqueeze(0)
        # no_ant = 1 - torch.sum(a - a * m > 0.5, dim=-1).bool().float()
        # coref_logits = torch.cat((coref_logits - coref_logits * m, no_ant.unsqueeze(-1)), dim=-1)
        no_ant = 1 - torch.sum(torch.sigmoid(coref_logits) > 0.5, dim=-1).bool().float()
        # [batch_size, max_k, max_k + 1]
        coref_logits = torch.cat((coref_logits, no_ant.unsqueeze(-1)), dim=-1)

        span_starts = span_starts.detach().cpu()
        span_ends = span_ends.detach().cpu()
        max_antecedents = coref_logits.argmax(axis=-1).detach().cpu()
        doc_indices = np.nonzero(max_antecedents < n_spans)[:, 0]
        # indices where antecedent is not null.
        mention_indices = np.nonzero(max_antecedents < n_spans)[:, 1]

        antecedent_indices = max_antecedents[max_antecedents < n_spans]
        span_indices = np.stack([span_starts.detach().cpu(), span_ends.detach().cpu()], axis=-1)

        mentions = span_indices[doc_indices, mention_indices]
        antecedents = span_indices[doc_indices, antecedent_indices]
        non_mentions = np.nonzero(max_antecedents == n_spans)[:, 1]

        sing_indices = np.zeros_like(len(np.setdiff1d(non_mentions, antecedent_indices)))
        singletons = span_indices[sing_indices, np.setdiff1d(non_mentions, antecedent_indices)]

        # mention_to_antecedent = np.stack([mentions, antecedents], axis=1)

        if len(mentions.shape) == 1 and len(antecedents.shape) == 1:
            mention_to_antecedent = np.stack([mentions, antecedents], axis=0)
        else:
            mention_to_antecedent = np.stack([mentions, antecedents], axis=1)

        if len(mentions.shape) == 1:
            mention_to_antecedent = [mention_to_antecedent]

        if len(singletons.shape) == 1:
            singletons = [singletons]

        return doc_indices, mention_to_antecedent, singletons

    def create_clusters(self, m2a, singletons, add):
        # Note: mention_to_antecedent is a numpy array
        if add != None:
            clusters = add
            mention_to_cluster = {m: i for i, c in enumerate(clusters) for m in c}
        else:
            clusters, mention_to_cluster = [], {}
        for mention, antecedent in m2a:
            mention, antecedent = tuple(mention), tuple(antecedent)
            if antecedent in mention_to_cluster:
                cluster_idx = mention_to_cluster[antecedent]
                if mention not in clusters[cluster_idx]:
                    clusters[cluster_idx].append(mention)
                    mention_to_cluster[mention] = cluster_idx
            elif mention in mention_to_cluster:
                cluster_idx = mention_to_cluster[mention]
                if antecedent not in clusters[cluster_idx]:
                    clusters[cluster_idx].append(antecedent)
                    mention_to_cluster[antecedent] = cluster_idx
            else:
                cluster_idx = len(clusters)
                mention_to_cluster[mention] = cluster_idx
                mention_to_cluster[antecedent] = cluster_idx
                clusters.append([antecedent, mention])

        clusters = [tuple(cluster) for cluster in clusters]
        # maybe order stuff?
        if len(singletons) != 0:
            clust = []
            while len(clusters) != 0 or len(singletons) != 0:
                if len(singletons) == 0:
                    clust.append(clusters[0])
                    clusters = clusters[1:]
                elif len(clusters) == 0:
                    clust.append(tuple([tuple(singletons[0])]))
                    singletons = singletons[1:]
                elif singletons[0][0] < sorted(clusters[0], key=lambda x: x[0])[0][0]:
                    clust.append(tuple([tuple(singletons[0])]))
                    singletons = singletons[1:]
                else:
                    clust.append(clusters[0])
                    clusters = clusters[1:]
            return clust
        return clusters

    def extract_clusters(self, gold_clusters):
        gold_clusters = [tuple(tuple(m) for m in cluster if (-1) not in m) for cluster in gold_clusters]
        gold_clusters = [cluster for cluster in gold_clusters if len(cluster) > 0]
        return gold_clusters

    def forward(
        self,
        stage,
        input_ids,
        attention_mask,
        eos_mask,
        gold_starts=None,
        gold_mentions=None,
        gold_clusters=None,
        tokens=None,
        subtoken_map=None,
        new_token_map=None,
        add=None,
        singletons=False,
        longdoc=None,
    ):

        loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)
        loss_dict = {}
        preds = {}
        if longdoc == None:
            last_hidden_states = self.encoder(input_ids=input_ids, attention_mask=attention_mask)[
                "last_hidden_state"
            ]  # B x S x TH

            lhs = last_hidden_states

            (
                start_idxs,
                mention_idxs,
                start_loss,
                mention_loss,
            ) = self.eos_mention_extraction(
                lhs=last_hidden_states,
                eos_mask=eos_mask,
                gold_mentions=gold_mentions,
                gold_starts=gold_starts,
                stage=stage,
            )

            loss_dict["start_loss"] = start_loss
            preds["start_idxs"] = [start.detach().cpu() for start in start_idxs]

            loss_dict["mention_loss"] = mention_loss
            preds["mention_idxs"] = [mention.detach().cpu() for mention in mention_idxs]

            loss = loss + start_loss + mention_loss

            if stage == "train":
                mention_idxs = (gold_mentions[0] == 1).nonzero(as_tuple=False)
            elif stage == "test" and gold_mentions != None:
                mention_idxs = (gold_mentions[0] == 1).nonzero(as_tuple=False)
            else:
                mention_idxs = mention_idxs[0]
            mention_start_idxs = mention_idxs[:, 0]
            mention_end_idxs = mention_idxs[:, 1]
            mentions_start_hidden_states = torch.index_select(lhs, 1, mention_start_idxs)
            mentions_end_hidden_states = torch.index_select(lhs, 1, mention_end_idxs)

        elif longdoc == "naive":
            mention_idxs, mentions_start_hidden_states, mentions_end_hidden_states = self.naive_long_doc_forward(
                stage, input_ids, attention_mask, eos_mask, gold_starts, gold_mentions
            )
            mention_idxs = mention_idxs[0]
        elif longdoc == "proper":
            pass
        else:
            raise Exception("unsupported")

        mention_start_idxs = mention_idxs[:, 0]
        mention_end_idxs = mention_idxs[:, 1]

        _, categories_masks = self._get_categories_labels(
            tokens, subtoken_map, new_token_map, mention_start_idxs, mention_end_idxs
        )

        coreference_loss, coreferences = self.mes_span_clustering(
            mentions_start_hidden_states,
            mentions_end_hidden_states,
            mention_start_idxs,
            mention_end_idxs,
            gold_clusters,
            stage,
            categories_masks,
            add,
            singletons,
        )

        loss = loss + coreference_loss
        loss_dict["coreference_loss"] = coreference_loss

        if stage != "train":
            preds["clusters"] = coreferences

        loss_dict["full_loss"] = loss
        output = {"pred_dict": preds, "loss_dict": loss_dict, "loss": loss}

        return output

    def naive_long_doc_forward(
        self,
        stage,
        input_ids,
        attention_mask,
        eos_mask,
        gold_starts=None,
        gold_mentions=None,
        max_seq_len=2000,
    ):
        length = input_ids.shape[1]
        slices_seq_index = [
            (eos_mask[0][step] == 1).nonzero(as_tuple=False)[-1][0].item() for step in range(max_seq_len, length, max_seq_len)
        ]
        star = (eos_mask[0][0] == 1).nonzero(as_tuple=False)[-1][0].item()
        if len(slices_seq_index) == 0 or slices_seq_index[-1] != length:
            slices_seq_index.append(length)
        prev = 0
        start_idxs = []
        mention_idxs = []
        index_start_hidden_states = []
        index_end_hidden_states = []
        for index in slices_seq_index:
            if prev != 0:
                index_input_ids = torch.cat((input_ids[0][0:star], input_ids[0][prev:index]), dim=0).unsqueeze(0)
                index_attention_mask = torch.cat((attention_mask[0][0:star], attention_mask[0][prev:index]), dim=0).unsqueeze(0)
                index_eos_mask = torch.cat((eos_mask[0][0:star], eos_mask[0][prev:index]), dim=0).unsqueeze(0)
                index_eos_mask = torch.cat((index_eos_mask[0][:, 0:star], index_eos_mask[0][:, prev:index]), dim=1).unsqueeze(0)
                if gold_mentions != None:
                    index_gold_mentions = torch.cat((gold_mentions[0][0:star], gold_mentions[0][prev:index]), dim=0).unsqueeze(0)
                    index_gold_mentions = torch.cat(
                        (index_gold_mentions[0][:, 0:star], index_gold_mentions[0][:, prev:index]), dim=1
                    ).unsqueeze(0)
                    index_gold_starts = torch.cat((gold_starts[0][0:star], input_ids[0][prev:index]), dim=0).unsqueeze(0)
                else:
                    index_gold_mentions = gold_mentions
                    index_gold_starts = gold_starts
            else:
                index_input_ids = input_ids[0][prev:index].unsqueeze(0)
                index_attention_mask = attention_mask[0][prev:index].unsqueeze(0)
                index_eos_mask = eos_mask[0][prev:index].unsqueeze(0)
                index_eos_mask = index_eos_mask[0][:, prev:index].unsqueeze(0)
                if gold_mentions != None:
                    index_gold_mentions = gold_mentions[0][prev:index].unsqueeze(0)
                    index_gold_mentions = index_gold_mentions[0][:, prev:index].unsqueeze(0)
                    index_gold_starts = gold_starts[0][prev:index].unsqueeze(0)
                else:
                    index_gold_mentions = gold_mentions
                    index_gold_starts = gold_starts

            index_last_hidden_states = self.encoder(input_ids=index_input_ids, attention_mask=index_attention_mask)[
                "last_hidden_state"
            ]  # B x S x TH

            (
                index_start_idxs,
                index_mention_idxs,
                _,
                _,
            ) = self.eos_mention_extraction(
                lhs=index_last_hidden_states,
                eos_mask=index_eos_mask,
                gold_mentions=index_gold_mentions,
                gold_starts=index_gold_starts,
                stage=stage,
            )

            if index_gold_mentions != None:
                index_mention_idxs = [(index_gold_mentions[0] == 1).nonzero(as_tuple=False)]
            if prev != 0:
                start_idxs.append(torch.add(index_start_idxs[0][index_start_idxs[0] > star], prev - star))
                mention_idxs.append(torch.add(index_mention_idxs[0][index_mention_idxs[0][:, 0] > star], prev - star))
                index_start_hidden_states.append(
                    torch.index_select(
                        index_last_hidden_states, 1, index_mention_idxs[0][index_mention_idxs[0][:, 0] > star][:, 0]
                    )
                )
                index_end_hidden_states.append(
                    torch.index_select(
                        index_last_hidden_states, 1, index_mention_idxs[0][index_mention_idxs[0][:, 0] > star][:, 1]
                    )
                )

            else:
                start_idxs.append(index_start_idxs[0])
                idxs = index_mention_idxs[0]
                mention_idxs.append(idxs)
                index_start_hidden_states.append(torch.index_select(index_last_hidden_states, 1, idxs[:, 0]))
                index_end_hidden_states.append(torch.index_select(index_last_hidden_states, 1, idxs[:, 1]))

            prev = index

        mentions_start_hidden_states = torch.cat(index_start_hidden_states, dim=1)
        mentions_end_hidden_states = torch.cat(index_end_hidden_states, dim=1)
        mention_idxs = [torch.cat(mention_idxs, dim=0)]
        return mention_idxs, mentions_start_hidden_states, mentions_end_hidden_states



import torch
import numpy as np
import random

from transformers import (
    AutoModel,
    AutoConfig,
    DistilBertForSequenceClassification,
    DistilBertConfig,
)

from maverick.common.util import *
from maverick.common.constants import *


class MentionClusterClassifier(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, mention_hidden_states, cluster_hidden_states, attention_mask, labels=None):
        # repreated tensor of mention_hs, to append in first position for each possible mention cluster pair
        repeated_mention_hs = mention_hidden_states.unsqueeze(0).repeat(cluster_hidden_states.shape[0], 1, 1)

        # mention cluste pairs by contatenating mention vectors to cluster padded matrix
        mention_cluster_pairs = torch.cat((repeated_mention_hs, cluster_hidden_states), dim=1)
        attention_mask = torch.cat(
            (
                torch.ones(cluster_hidden_states.shape[0], 1, device=self.model.device),
                attention_mask,
            ),
            dim=1,
        )

        logits = self.model(inputs_embeds=mention_cluster_pairs, attention_mask=attention_mask).logits

        loss = None
        if labels is not None:
            loss = torch.nn.functional.binary_cross_entropy_with_logits(logits, labels.unsqueeze(1).to(self.model.device))
        return loss, logits


class Maverick_incr(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        # document transformer encoder
        self.encoder_hf_model_name = kwargs["huggingface_model_name"]
        self.encoder = AutoModel.from_pretrained(self.encoder_hf_model_name)
        self.encoder_config = AutoConfig.from_pretrained(self.encoder_hf_model_name)
        self.encoder.resize_token_embeddings(self.encoder.embeddings.word_embeddings.num_embeddings + 3)

        # freeze
        if kwargs["freeze_encoder"]:
            for param in self.encoder.parameters():
                param.requires_grad = False

        # type of representation layer in 'Linear, FC, LSTM-left, LSTM-right, Conv1d'
        self.representation_layer_type = "FC"  # fullyconnected
        # span hidden dimension
        self.token_hidden_size = self.encoder_config.hidden_size

        # if span representation method is to concatenate start and end, a mention hidden size will be 2*token_hidden_size
        self.mention_hidden_size = self.token_hidden_size * 2

        # incremental transformer classifier
        self.incremental_model_hidden_size = kwargs.get("incremental_model_hidden_size", 384)  # 768/2
        self.incremental_model_num_layers = kwargs.get("incremental_model_num_layers", 1)
        self.incremental_model_config = DistilBertConfig(num_labels=1, hidden_size=self.incremental_model_hidden_size)
        self.incremental_model = DistilBertForSequenceClassification(self.incremental_model_config).to(self.encoder.device)
        self.incremental_model.distilbert.transformer.layer = self.incremental_model.distilbert.transformer.layer[
            : self.incremental_model_num_layers
        ]
        self.incremental_model.distilbert.embeddings.word_embeddings = None
        self.incremental_transformer = MentionClusterClassifier(model=self.incremental_model)

        # encodes mentions for incremental clustering
        self.incremental_span_encoder = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.mention_hidden_size,
            output_dim=self.incremental_model_hidden_size,
            hidden_dim=int(self.mention_hidden_size / 2),
        )

        # mention extraction layers
        # representation of start token
        self.start_token_representation = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        # representation of end token
        self.end_token_representation = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        # models probability to be the start of a mention
        self.start_token_classifier = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=1,
            hidden_dim=self.token_hidden_size,
        )

        # model mention probability from start and end representations
        self.start_end_classifier = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.mention_hidden_size,
            output_dim=1,
            hidden_dim=self.token_hidden_size,
        )

    # takes last_hidden_states, eos_mask, ground truth and stage
    def squad_mention_extraction(self, lhs, eos_mask, gold_mentions, gold_starts, stage):
        start_idxs = []
        mention_idxs = []
        start_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)
        mention_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)

        for bidx in range(0, lhs.shape[0]):
            lhs_batch = lhs[bidx]  # SEQ_LEN X HIDD_DIM
            eos_mask_batch = eos_mask[bidx]  # SEQ_LEN X SEQ_LEN

            # compute start logits
            start_logits_batch = self.start_token_classifier(lhs_batch).squeeze(-1)  # SEQ_LEN

            if gold_starts != None:
                loss = torch.nn.functional.binary_cross_entropy_with_logits(start_logits_batch, gold_starts[bidx])

                # accumulate loss
                start_loss = start_loss + loss

            # compute start positions
            start_idxs_batch = ((torch.sigmoid(start_logits_batch) > 0.5)).nonzero(as_tuple=False).squeeze(-1)

            start_idxs.append(start_idxs_batch.detach().clone())
            # in training, use gold starts to learn to extract mentions, inference use predicted ones
            if stage == "train":
                start_idxs_batch = (
                    ((torch.sigmoid(gold_starts[bidx]) > 0.5)).nonzero(as_tuple=False).squeeze(-1)
                )  # NUM_GOLD_STARTS

            # contains all possible start end indices pairs, i.e. for all starts, all possible ends looking at EOS index
            possibles_start_end_idxs = (eos_mask_batch[start_idxs_batch] == 1).nonzero(as_tuple=False)  # STARTS x 2

            # this is to have reference respect to original positions
            possibles_start_end_idxs[:, 0] = start_idxs_batch[possibles_start_end_idxs[:, 0]]

            possible_start_idxs = possibles_start_end_idxs[:, 0]
            possible_end_idxs = possibles_start_end_idxs[:, 1]

            # extract start and end hidden states
            starts_hidden_states = lhs_batch[possible_end_idxs]  # start
            ends_hidden_states = lhs_batch[possible_start_idxs]  # end

            # concatenation of start to end representations created using a representation layer
            s2e_representations = torch.cat(
                (
                    self.start_token_representation(starts_hidden_states),
                    self.end_token_representation(ends_hidden_states),
                ),
                dim=-1,
            )

            # classification of mentions
            s2e_logits = self.start_end_classifier(s2e_representations).squeeze(-1)

            # mention_start_idxs and mention_end_idxs
            mention_idxs.append(possibles_start_end_idxs[torch.sigmoid(s2e_logits) > 0.5].detach().clone())

            if s2e_logits.shape[0] != 0:
                if gold_mentions != None:
                    mention_loss_batch = torch.nn.functional.binary_cross_entropy_with_logits(
                        s2e_logits,
                        gold_mentions[bidx][possible_start_idxs, possible_end_idxs],
                    )
                    mention_loss = mention_loss + mention_loss_batch

        return (start_idxs, mention_idxs, start_loss, mention_loss)

    def incremental_span_clustering(self, mentions_hidden_states, mentions_idxs, gold_clusters, stage):
        pred_cluster_idxs = []  # cluster_idxs = list of list of tuple of offsets (also output) up to mention_idx
        if gold_clusters != None:
            gold_cluster_idxs = unpad_gold_clusters(gold_clusters)  # gold_cluster_idxs, but padded

        coreference_loss = torch.tensor([0.0], requires_grad=True, device=self.incremental_model.device)
        mentions_hidden_states = mentions_hidden_states[0]
        idx_to_hs = dict(zip([tuple(m) for m in mentions_idxs.tolist()], mentions_hidden_states))

        # for each mention
        for idx, (
            mention_hidden_states,
            (mention_start_idx, mention_end_idx),
        ) in enumerate(zip(mentions_hidden_states, mentions_idxs)):
            if idx == 0:
                # if first create singleton cluster
                pred_cluster_idxs.append([(mention_start_idx.item(), mention_end_idx.item())])
            else:
                if stage == "train":
                    # if we are in training, retrieve use gold cluster idx to induce loss.
                    cluster_idx, labels = self.new_cluster_idxs_labels(
                        (mention_start_idx, mention_end_idx), gold_cluster_idxs
                    )  # can be used using only tensors
                else:
                    cluster_idx, labels = pred_cluster_idxs, None

                # get cluster padded matrix matrix and attention mask (excludes padding)
                cluster_hs, cluster_am = self.get_cluster_states_matrix(idx_to_hs, cluster_idx, stage)

                # produce logits for each possible cluster mention pair
                mention_cluster_loss, logits = self.incremental_transformer(
                    mention_hidden_states=mention_hidden_states,
                    cluster_hidden_states=cluster_hs,
                    attention_mask=cluster_am,
                    labels=labels,
                )

                if mention_cluster_loss != None:
                    coreference_loss = coreference_loss + mention_cluster_loss

                if stage != "train":
                    # only in inference
                    num_possible_clustering = torch.sum(torch.sigmoid(logits) > 0.5, dim=0).bool().float()

                    if num_possible_clustering == 0:
                        # if no clustering, create new singleton cluster
                        pred_cluster_idxs.append([(mention_start_idx.item(), mention_end_idx.item())])
                    else:
                        # otherwise, take most probabile clustering predicted by the model and assign this mention to that cluster
                        assigned_idx = logits.argmax(axis=0).detach().cpu()
                        pred_cluster_idxs[assigned_idx.item()].append((mention_start_idx.item(), mention_end_idx.item()))
        # normalize loss debug
        if gold_clusters != None:
            coreference_loss = coreference_loss / (mentions_hidden_states.shape[0] if mentions_hidden_states.shape[0] != 0 else 1)

            # coreference_loss = coreference_loss / (len(gold_cluster_idxs) if len(gold_cluster_idxs) != 0 else 1)
        coreferences_pred = [tuple(item) for item in pred_cluster_idxs]  # if len(item) > 1]
        return coreference_loss, coreferences_pred

    def get_cluster_states_matrix(self, idx_to_hs, cluster_idxs, stage):
        # create padded matrix of encoded mentions
        max_length = max([len(x) for x in cluster_idxs])
        if stage == "train":
            max_length = max_length if max_length < 31 else 30
        forward_matrix = torch.zeros(
            (len(cluster_idxs), max_length, self.incremental_model_hidden_size),
            device=self.encoder.device,
        )
        forward_am = torch.zeros((len(cluster_idxs), max_length), device=self.encoder.device)

        for cluster_idx, span_idxs in enumerate(cluster_idxs):
            if stage == "train":
                if len(span_idxs) > 30:
                    span_idxs = sorted(span_idxs)
                    new_idxs = [span_idxs[0]]
                    new_idxs.extend(random.sample(span_idxs, 28))
                    new_idxs.append(span_idxs[-1])
                    span_idxs = new_idxs

            hs = torch.stack([idx_to_hs[span_idx] for span_idx in span_idxs])

            forward_matrix[cluster_idx][: hs.shape[0]] = hs
            forward_am[cluster_idx][: hs.shape[0]] = torch.ones((hs.shape[0]), device=self.encoder.device)

        return forward_matrix, forward_am

    # takes the index of the mention (mention_start, mention_end) and gold coreferences, returns filtered indices (up to mention idx) and labels
    def new_cluster_idxs_labels(self, mention_idxs, gold_coreference_idxs):
        res_coreference_idxs = []
        # list of length number of clusters in gold, and 1.0 where the mention is laying
        labels = [
            1.0 if (mention_idxs[0].item(), mention_idxs[1].item()) in span_idx else 0.0 for span_idx in gold_coreference_idxs
        ]
        # filter cluster up to the mention you are evaluating
        for cluster_idxs in gold_coreference_idxs:
            idxs = []
            for span_idx in cluster_idxs:
                # if span is antecedent to current mention, stay in possible clusters
                if span_idx[0] < mention_idxs[0].item() or (
                    span_idx[0] == mention_idxs[0].item() and span_idx[1] < mention_idxs[1].item()
                ):
                    idxs.append((span_idx[0], span_idx[1]))
            # idxs = sorted(idxs, reverse=True)
            res_coreference_idxs.append(idxs)

        labels = torch.tensor(
            [lab for lab, idx in zip(labels, res_coreference_idxs) if len(idx) != 0],
            device=self.encoder.device,
        )
        res_coreference_idxs = [idx for idx in res_coreference_idxs if len(idx) != 0]
        return res_coreference_idxs, labels

    def forward(
        self,
        stage,
        input_ids,
        attention_mask,
        eos_mask,
        gold_starts=None,
        gold_mentions=None,
        gold_clusters=None,
    ):
        last_hidden_states = self.encoder(input_ids=input_ids, attention_mask=attention_mask)["last_hidden_state"]  # B x S x TH

        lhs = last_hidden_states

        loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)
        loss_dict = {}
        preds = {}

        (
            start_idxs,
            mention_idxs,
            start_loss,
            mention_loss,
        ) = self.squad_mention_extraction(
            lhs=last_hidden_states,
            eos_mask=eos_mask,
            gold_mentions=gold_mentions,
            gold_starts=gold_starts,
            stage=stage,
        )

        loss_dict["start_loss"] = start_loss
        preds["start_idxs"] = [start.detach().cpu() for start in start_idxs]

        loss_dict["mention_loss"] = mention_loss
        preds["mention_idxs"] = [mention.detach().cpu() for mention in mention_idxs]

        loss = loss + start_loss + mention_loss

        if stage == "train":
            mention_idxs = (gold_mentions[0] == 1).nonzero(as_tuple=False)
        else:
            mention_idxs = mention_idxs[0]

        mention_start_idxs = mention_idxs[:, 0]
        mention_end_idxs = mention_idxs[:, 1]

        mentions_start_hidden_states = torch.index_select(lhs, 1, mention_start_idxs)
        mentions_end_hidden_states = torch.index_select(lhs, 1, mention_end_idxs)

        mentions_hidden_states = torch.cat((mentions_start_hidden_states, mentions_end_hidden_states), dim=2)

        mentions_hidden_states = self.incremental_span_encoder(mentions_hidden_states)

        coreference_loss, coreferences = self.incremental_span_clustering(
            mentions_hidden_states, mention_idxs, gold_clusters, stage
        )

        loss = loss + coreference_loss
        loss_dict["coreference_loss"] = coreference_loss

        if stage != "train":
            preds["clusters"] = coreferences

        loss_dict["full_loss"] = loss
        output = {"pred_dict": preds, "loss_dict": loss_dict, "loss": loss}

        return output



import torch
import numpy as np

from transformers import AutoModel, AutoConfig

from maverick.common.util import *
from maverick.common.constants import *


class Maverick_s2e(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        # document transformer encoder
        self.encoder_hf_model_name = kwargs["huggingface_model_name"]
        self.encoder = AutoModel.from_pretrained(self.encoder_hf_model_name)
        self.encoder_config = AutoConfig.from_pretrained(self.encoder_hf_model_name)
        # self.encoder_config.attention_window = 1024
        self.encoder.resize_token_embeddings(self.encoder.embeddings.word_embeddings.num_embeddings + 3)

        # freeze
        if kwargs["freeze_encoder"]:
            for param in self.encoder.parameters():
                param.requires_grad = False

        # span representation, now is concat_start_end
        self.span_representation = kwargs["span_representation"]
        # type of representation layer in 'Linear, FC, LSTM-left, LSTM-right, Conv1d'
        self.representation_layer_type = "FC"  # fullyconnected
        # span hidden dimension
        self.token_hidden_size = self.encoder_config.hidden_size

        # if span representation method is to concatenate start and end, a mention hidden size will be 2*token_hidden_size
        if self.span_representation == "concat_start_end":
            self.mention_hidden_size = self.token_hidden_size * 2

        self.antecedent_s2s_classifier = RepresentationLayer(
            type=self.representation_layer_type,  # fullyconnected
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )
        self.antecedent_e2e_classifier = RepresentationLayer(
            type=self.representation_layer_type,  # fullyconnected
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        self.antecedent_s2e_classifier = RepresentationLayer(
            type=self.representation_layer_type,  # fullyconnected
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )
        self.antecedent_e2s_classifier = RepresentationLayer(
            type=self.representation_layer_type,  # fullyconnected
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        # mention extraction layers
        # representation of start token
        self.start_token_representation = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        # representation of end token
        self.end_token_representation = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=self.token_hidden_size,
            hidden_dim=self.token_hidden_size,
        )

        # models probability to be the start of a mention
        self.start_token_classifier = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.token_hidden_size,
            output_dim=1,
            hidden_dim=self.token_hidden_size,
        )

        # model mention probability from start and end representations
        self.start_end_classifier = RepresentationLayer(
            type=self.representation_layer_type,
            input_dim=self.mention_hidden_size,
            output_dim=1,
            hidden_dim=self.token_hidden_size,
        )

    # takes last_hidden_states, eos_mask, ground truth and stage
    # check if mention_mask deletes some good gold mentions!
    def squad_mention_extraction(self, lhs, eos_mask, gold_mentions, gold_starts, stage):
        start_idxs = []
        mention_idxs = []
        start_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)
        mention_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)

        for bidx in range(0, lhs.shape[0]):
            lhs_batch = lhs[bidx]  # SEQ_LEN X HIDD_DIM
            eos_mask_batch = eos_mask[bidx]  # SEQ_LEN X SEQ_LEN

            # compute start logits
            start_logits_batch = self.start_token_classifier(lhs_batch).squeeze(-1)  # SEQ_LEN

            if gold_starts != None:
                loss = torch.nn.functional.binary_cross_entropy_with_logits(start_logits_batch, gold_starts[bidx])

                # accumulate loss
                start_loss = start_loss + loss

            # compute start positions
            start_idxs_batch = ((torch.sigmoid(start_logits_batch) > 0.5)).nonzero(as_tuple=False).squeeze(-1)

            start_idxs.append(start_idxs_batch.detach().clone())
            # in training, use gold starts to learn to extract mentions, inference use predicted ones
            if stage == "train":
                start_idxs_batch = (
                    ((torch.sigmoid(gold_starts[bidx]) > 0.5)).nonzero(as_tuple=False).squeeze(-1)
                )  # NUM_GOLD_STARTS

            # contains all possible start end indices pairs, i.e. for all starts, all possible ends looking at EOS index
            possibles_start_end_idxs = (eos_mask_batch[start_idxs_batch] == 1).nonzero(as_tuple=False)  # STARTS x 2

            # this is to have reference respect to original positions
            possibles_start_end_idxs[:, 0] = start_idxs_batch[possibles_start_end_idxs[:, 0]]

            possible_start_idxs = possibles_start_end_idxs[:, 0]
            possible_end_idxs = possibles_start_end_idxs[:, 1]

            # extract start and end hidden states
            starts_hidden_states = lhs_batch[possible_end_idxs]  # start
            ends_hidden_states = lhs_batch[possible_start_idxs]  # end

            # concatenation of start to end representations created using a representation layer
            s2e_representations = torch.cat(
                (
                    self.start_token_representation(starts_hidden_states),
                    self.end_token_representation(ends_hidden_states),
                ),
                dim=-1,
            )

            # classification of mentions
            s2e_logits = self.start_end_classifier(s2e_representations).squeeze(-1)

            # mention_start_idxs and mention_end_idxs
            mention_idxs.append(possibles_start_end_idxs[torch.sigmoid(s2e_logits) > 0.5].detach().clone())

            if s2e_logits.shape[0] != 0:
                if gold_mentions != None:
                    mention_loss_batch = torch.nn.functional.binary_cross_entropy_with_logits(
                        s2e_logits,
                        gold_mentions[bidx][possible_start_idxs, possible_end_idxs],
                    )
                    mention_loss = mention_loss + mention_loss_batch

        return (start_idxs, mention_idxs, start_loss, mention_loss)

    def create_mention_to_antecedent_singletons(self, span_starts, span_ends, coref_logits):
        span_starts = span_starts.unsqueeze(0)
        span_ends = span_ends.unsqueeze(0)
        bs, n_spans, _ = coref_logits.shape

        no_ant = 1 - torch.sum(torch.sigmoid(coref_logits) > 0.5, dim=-1).bool().float()
        # [batch_size, max_k, max_k + 1]
        coref_logits = torch.cat((coref_logits, no_ant.unsqueeze(-1)), dim=-1)

        span_starts = span_starts.detach().cpu()
        span_ends = span_ends.detach().cpu()
        max_antecedents = coref_logits.argmax(axis=-1).detach().cpu()
        doc_indices = np.nonzero(max_antecedents < n_spans)[:, 0]
        # indices where antecedent is not null.
        mention_indices = np.nonzero(max_antecedents < n_spans)[:, 1]

        antecedent_indices = max_antecedents[max_antecedents < n_spans]
        span_indices = np.stack([span_starts.detach().cpu(), span_ends.detach().cpu()], axis=-1)

        mentions = span_indices[doc_indices, mention_indices]
        antecedents = span_indices[doc_indices, antecedent_indices]
        non_mentions = np.nonzero(max_antecedents == n_spans)[:, 1]

        sing_indices = np.zeros_like(len(np.setdiff1d(non_mentions, antecedent_indices)))
        singletons = span_indices[sing_indices, np.setdiff1d(non_mentions, antecedent_indices)]

        # mention_to_antecedent = np.stack([mentions, antecedents], axis=1)

        if len(mentions.shape) == 1 and len(antecedents.shape) == 1:
            mention_to_antecedent = np.stack([mentions, antecedents], axis=0)
        else:
            mention_to_antecedent = np.stack([mentions, antecedents], axis=1)

        if len(mentions.shape) == 1:
            mention_to_antecedent = [mention_to_antecedent]

        if len(singletons.shape) == 1:
            singletons = [singletons]

        return doc_indices, mention_to_antecedent, singletons

    def _get_cluster_labels_after_pruning(self, span_starts, span_ends, all_clusters):
        """
        :param span_starts: [batch_size, max_k]
        :param span_ends: [batch_size, max_k]
        :param all_clusters: [batch_size, max_cluster_size, max_clusters_num, 2]
        :return: [batch_size, max_k, max_k + 1] - [b, i, j] == 1 if i is antecedent of j
        """
        span_starts = span_starts.unsqueeze(0)
        span_ends = span_ends.unsqueeze(0)
        batch_size, max_k = span_starts.size()
        new_cluster_labels = torch.zeros((batch_size, max_k, max_k), device="cpu")
        all_clusters_cpu = all_clusters.cpu().numpy()
        for b, (starts, ends, gold_clusters) in enumerate(
            zip(span_starts.cpu().tolist(), span_ends.cpu().tolist(), all_clusters_cpu)
        ):
            gold_clusters = self.extract_clusters(gold_clusters)
            mention_to_gold_clusters = extract_mentions_to_clusters(gold_clusters)
            gold_mentions = set(mention_to_gold_clusters.keys())
            for i, (start, end) in enumerate(zip(starts, ends)):
                if (start, end) not in gold_mentions:
                    continue
                for j, (a_start, a_end) in enumerate(list(zip(starts, ends))[:i]):
                    if (a_start, a_end) in mention_to_gold_clusters[(start, end)]:
                        new_cluster_labels[b, i, j] = 1
        new_cluster_labels = new_cluster_labels.to(self.encoder.device)
        return new_cluster_labels

    def s2e_mention_cluster(self, mention_start_reps, mention_end_reps, mention_start_idxs, mention_end_idxs, gold, stage):
        coref_logits = self._calc_coref_logits(mention_start_reps, mention_end_reps)
        coreference_loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)

        coref_logits = torch.stack([matrix.tril().fill_diagonal_(0) for matrix in coref_logits])
        if stage == "train":
            labels = self._get_cluster_labels_after_pruning(mention_start_idxs, mention_end_idxs, gold)
            coreference_loss = torch.nn.functional.binary_cross_entropy_with_logits(coref_logits, labels)

        doc, m2a, singletons = self.create_mention_to_antecedent_singletons(mention_start_idxs, mention_end_idxs, coref_logits)
        coreferences = self.create_clusters(m2a, singletons)
        return coreference_loss, coreferences

    def create_clusters(self, m2a, singletons):
        # Note: mention_to_antecedent is a numpy array

        clusters, mention_to_cluster = [], {}
        for mention, antecedent in m2a:
            mention, antecedent = tuple(mention), tuple(antecedent)
            if antecedent in mention_to_cluster:
                cluster_idx = mention_to_cluster[antecedent]
                if mention not in clusters[cluster_idx]:
                    clusters[cluster_idx].append(mention)
                    mention_to_cluster[mention] = cluster_idx
            elif mention in mention_to_cluster:
                cluster_idx = mention_to_cluster[mention]
                if antecedent not in clusters[cluster_idx]:
                    clusters[cluster_idx].append(antecedent)
                    mention_to_cluster[antecedent] = cluster_idx
            else:
                cluster_idx = len(clusters)
                mention_to_cluster[mention] = cluster_idx
                mention_to_cluster[antecedent] = cluster_idx
                clusters.append([antecedent, mention])

        clusters = [tuple(cluster) for cluster in clusters]
        if len(singletons) != 0:
            clust = []
            while len(clusters) != 0 or len(singletons) != 0:
                if len(singletons) == 0:
                    clust.append(clusters[0])
                    clusters = clusters[1:]
                elif len(clusters) == 0:
                    clust.append(tuple([tuple(singletons[0])]))
                    singletons = singletons[1:]
                elif singletons[0][0] < sorted(clusters[0], key=lambda x: x[0])[0][0]:
                    clust.append(tuple([tuple(singletons[0])]))
                    singletons = singletons[1:]
                else:
                    clust.append(clusters[0])
                    clusters = clusters[1:]
            return clust
        return clusters

    def _calc_coref_logits(self, top_k_start_coref_reps, top_k_end_coref_reps):
        # s2s
        temp = self.antecedent_s2s_classifier(top_k_start_coref_reps)  # [batch_size, max_k, dim]
        top_k_s2s_coref_logits = torch.matmul(temp, top_k_start_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # e2e
        temp = self.antecedent_e2e_classifier(top_k_end_coref_reps)  # [batch_size, max_k, dim]
        top_k_e2e_coref_logits = torch.matmul(temp, top_k_end_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # s2e
        temp = self.antecedent_s2e_classifier(top_k_start_coref_reps)  # [batch_size, max_k, dim]
        top_k_s2e_coref_logits = torch.matmul(temp, top_k_end_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # e2s
        temp = self.antecedent_e2s_classifier(top_k_end_coref_reps)  # [batch_size, max_k, dim]
        top_k_e2s_coref_logits = torch.matmul(temp, top_k_start_coref_reps.permute([0, 2, 1]))  # [batch_size, max_k, max_k]

        # sum all terms
        coref_logits = (
            top_k_s2e_coref_logits + top_k_e2s_coref_logits + top_k_s2s_coref_logits + top_k_e2e_coref_logits
        )  # [batch_size, max_k, max_k]
        return coref_logits

    def extract_clusters(gold_clusters):
        gold_clusters = [tuple(tuple(m) for m in cluster if (-1) not in m) for cluster in gold_clusters]
        gold_clusters = [cluster for cluster in gold_clusters if len(cluster) > 0]
        return gold_clusters

    def forward(
        self,
        stage,
        input_ids,
        attention_mask,
        eos_mask,
        gold_starts=None,
        gold_mentions=None,
        gold_clusters=None,
    ):
        last_hidden_states = self.encoder(input_ids=input_ids, attention_mask=attention_mask)["last_hidden_state"]  # B x S x TH

        lhs = last_hidden_states

        loss = torch.tensor([0.0], requires_grad=True, device=self.encoder.device)
        loss_dict = {}
        preds = {}

        (
            start_idxs,
            mention_idxs,
            start_loss,
            mention_loss,
        ) = self.squad_mention_extraction(
            lhs=last_hidden_states,
            eos_mask=eos_mask,
            gold_mentions=gold_mentions,
            gold_starts=gold_starts,
            stage=stage,
        )

        loss_dict["start_loss"] = start_loss
        preds["start_idxs"] = [start.detach().cpu() for start in start_idxs]

        loss_dict["mention_loss"] = mention_loss
        preds["mention_idxs"] = [mention.detach().cpu() for mention in mention_idxs]

        loss = loss + start_loss + mention_loss

        if stage == "train":
            mention_idxs = (gold_mentions[0] == 1).nonzero(as_tuple=False)
        else:
            mention_idxs = mention_idxs[0]

        mention_start_idxs = mention_idxs[:, 0]
        mention_end_idxs = mention_idxs[:, 1]

        mentions_start_hidden_states = torch.index_select(lhs, 1, mention_start_idxs)
        mentions_end_hidden_states = torch.index_select(lhs, 1, mention_end_idxs)

        coreference_loss, coreferences = self.s2e_mention_cluster(
            mentions_start_hidden_states,
            mentions_end_hidden_states,
            mention_start_idxs,
            mention_end_idxs,
            gold_clusters,
            stage,
        )

        loss = loss + coreference_loss
        loss_dict["coreference_loss"] = coreference_loss

        if stage != "train":
            preds["clusters"] = coreferences

        loss_dict["full_loss"] = loss
        output = {"pred_dict": preds, "loss_dict": loss_dict, "loss": loss}

        return output



import os
import hydra
import omegaconf
import pytorch_lightning as pl
import torch

from pathlib import Path
from typing import Optional
from omegaconf import OmegaConf
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import (
    EarlyStopping,
    ModelCheckpoint,
    LearningRateMonitor,
)
from pytorch_lightning.callbacks.progress.rich_progress import RichProgressBar
from pytorch_lightning.loggers import WandbLogger
from rich.console import Console

from maverick.data.pl_data_modules import BasePLDataModule
from maverick.models.pl_modules import BasePLModule

torch.set_printoptions(edgeitems=100)


def train(conf: omegaconf.DictConfig) -> None:
    # fancy logger
    console = Console()
    # reproducibility
    pl.seed_everything(conf.train.seed)
    set_determinism_the_old_way(conf.train.pl_trainer.deterministic)
    conf.train.pl_trainer.deterministic = True

    console.log(f"Starting training for [bold cyan]{conf.train.model_name}[/bold cyan] model")
    if conf.train.pl_trainer.fast_dev_run:
        console.log(f"Debug mode {conf.train.pl_trainer.fast_dev_run}. Forcing debugger configuration")
        # Debuggers don't like GPUs nor multiprocessing
        conf.train.pl_trainer.accelerator = "cpu"

        conf.train.pl_trainer.precision = 32
        conf.data.datamodule.num_workers = {k: 0 for k in conf.data.datamodule.num_workers}
        # Switch wandb to offline mode to prevent online logging
        conf.logging.log = None
        # remove model checkpoint callback
        conf.train.model_checkpoint_callback = None

    # data module declaration
    console.log(f"Instantiating the Data Module")
    pl_data_module: BasePLDataModule = hydra.utils.instantiate(conf.data.datamodule, _recursive_=False)
    # force setup to get labels initialized for the model
    pl_data_module.prepare_data()
    pl_data_module.setup("fit")

    # main module declaration
    console.log(f"Instantiating the Model")

    pl_module: BasePLModule = hydra.utils.instantiate(conf.model.module, _recursive_=False)

    # pl_module = BasePLModule.load_from_checkpoint(conf.evaluation.checkpoint, _recursive_=False, map_location="cuda:0")
    experiment_logger: Optional[WandbLogger] = None
    experiment_path: Optional[Path] = None
    if conf.logging.log:
        console.log(f"Instantiating Wandb Logger")
        experiment_logger = hydra.utils.instantiate(conf.logging.wandb_arg)
        experiment_logger.watch(pl_module, **conf.logging.watch)
        experiment_path = Path(experiment_logger.experiment.dir)
        # Store the YaML config separately into the wandb dir
        yaml_conf: str = OmegaConf.to_yaml(cfg=conf)
        (experiment_path / "hparams.yaml").write_text(yaml_conf)

        # callbacks declaration
    callbacks_store = [RichProgressBar()]

    if conf.train.early_stopping_callback is not None:
        early_stopping_callback: EarlyStopping = hydra.utils.instantiate(conf.train.early_stopping_callback)
        callbacks_store.append(early_stopping_callback)

    if conf.train.model_checkpoint_callback is not None:
        model_checkpoint_callback: ModelCheckpoint = hydra.utils.instantiate(
            conf.train.model_checkpoint_callback,
            dirpath=experiment_path / "checkpoints",
        )
        callbacks_store.append(model_checkpoint_callback)

    if conf.train.learning_rate_callback is not None and not conf.train.pl_trainer.fast_dev_run:
        lr: LearningRateMonitor = hydra.utils.instantiate(conf.train.learning_rate_callback)
        callbacks_store.append(lr)
    # trainer
    console.log(f"Instantiating the Trainer")
    trainer: Trainer = hydra.utils.instantiate(conf.train.pl_trainer, callbacks=callbacks_store, logger=experiment_logger)

    # module fit
    trainer.fit(pl_module, datamodule=pl_data_module)

    # module test
    trainer.test(pl_module, datamodule=pl_data_module)


def set_determinism_the_old_way(deterministic: bool):
    # determinism for cudnn
    torch.backends.cudnn.deterministic = deterministic
    if deterministic:
        # fixing non-deterministic part of horovod
        # https://github.com/PyTorchLightning/pytorch-lightning/pull/1572/files#r420279383
        os.environ["HOROVOD_FUSION_THRESHOLD"] = str(0)


@hydra.main(config_path="../conf", config_name="root")
def main(conf: omegaconf.DictConfig):
    print(OmegaConf.to_yaml(conf))
    train(conf)


if __name__ == "__main__":
    main()



import json
import hydra
import subprocess
import torch
from omegaconf import omegaconf
from maverick.common.util import *
from maverick.common.metrics import *
from tqdm import tqdm
from data.pl_data_modules import BasePLDataModule
from models.pl_modules import BasePLModule
from maverick.utils.loggingl import get_console_logger

logger = get_console_logger()


def jsonlines_to_html(jsonlines_input_name, output):
    cwd = str(hydra.utils.get_original_cwd())
    subprocess.call(
        "python3 "
        + cwd
        + "/maverick/utils/corefconversion/jsonlines2text.py "
        + cwd
        + "/"
        + jsonlines_input_name
        + " -i -o "
        + cwd
        + "/experiments/"
        + output
        + ".html --sing-color"
        ' "black" --cm "common"',
        shell=True,
    )


@torch.no_grad()
def evaluate(conf: omegaconf.DictConfig):
    device = conf.evaluation.device
    hydra.utils.log.info("Using {} as device".format(device))
    pl_data_module: BasePLDataModule = hydra.utils.instantiate(conf.data.datamodule, _recursive_=False)

    pl_data_module.prepare_data()
    pl_data_module.setup("test")

    # jsonlines_to_html(pl_data_module.test_dataloader().dataset.path, "test")
    logger.log(f"Instantiating the Model from {conf.evaluation.checkpoint}")
    model = BasePLModule.load_from_checkpoint(conf.evaluation.checkpoint, _recursive_=False, map_location=device)
    if "gap" not in pl_data_module.test_dataloader().dataset.path:
        gold = []
        info = []
        with open(hydra.utils.get_original_cwd() + "/" + pl_data_module.test_dataloader().dataset.path, "r") as f:
            for line in f.readlines():
                doc = json.loads(line)
                if "sentences" in doc:
                    info.append({"doc_key": doc["doc_key"], "sentences": doc["sentences"]})
                if "text" in doc:
                    info.append({"doc_key": doc["doc_key"], "text": doc["text"]})
                clusters = []
                if "clusters" in doc:
                    for cluster in doc["clusters"]:
                        clusters.append(tuple([(m[0], m[1]) for m in cluster]))
                gold.append(clusters)

        mention_to_gold_clusters = [extract_mentions_to_clusters([tuple(g) for g in gold_element]) for gold_element in gold]

        predictions = model_predictions_with_dataloader(model, pl_data_module.test_dataloader(), device)
        mention_to_predicted_clusters = [extract_mentions_to_clusters(p) for p in predictions]

        print(evaluate_coref_scores(predictions, gold, mention_to_predicted_clusters, mention_to_gold_clusters))

        with open(hydra.utils.get_original_cwd() + "/experiments/output.jsonlines", "w") as f:
            for pred, infos in zip(predictions, info):
                infos["clusters"] = pred
                f.write(json.dumps(infos) + "\n")

                # jsonlines_to_html("experiments/output.jsonlines", "output")
    else:
        predictions = model_predictions_with_dataloader(model, pl_data_module.test_dataloader(), device)
        with open("data/gap/gap-test-ontoformat.jsonl", "r") as fr:
            with open("data/gap/gap-test-output.tsv", "w") as fw:
                for line, pred in zip(fr.readlines(), predictions):
                    doc = json.loads(line)
                    # flattened = flatten(doc["sentences"])
                    A, B = A_and_B_corefs(pred, doc["gap_pronoun_offsets"], doc["gap_A_offsets"], doc["gap_B_offsets"])
                    fw.write(doc["doc_key"] + "\t" + str(A) + "\t" + str(B) + "\n")
    return


def A_and_B_corefs(clusters, pronoun_offsets, A_offsets, B_offsets):
    A_Coref = False
    B_Coref = False
    pronoun_in_clusters = find_offsets_in_clusters(clusters, pronoun_offsets)
    A_in_clusters = find_offsets_in_clusters(clusters, A_offsets)
    B_in_clusters = find_offsets_in_clusters(clusters, B_offsets)
    # pronoun is in clusters
    if pronoun_in_clusters != None:
        # A is in clusters
        if A_in_clusters != None and A_in_clusters != B_in_clusters:
            A_Coref = candidate_pronoun_clustered(clusters, pronoun_in_clusters, A_in_clusters)
        if B_in_clusters != None and B_in_clusters != A_in_clusters:
            B_Coref = candidate_pronoun_clustered(clusters, pronoun_in_clusters, B_in_clusters)
    return A_Coref, B_Coref


def candidate_pronoun_clustered(clusters, pronoun, candidate):
    clustered = False
    for cluster in clusters:
        if pronoun in cluster and candidate in cluster:
            clustered = True
    return clustered


def find_offsets_in_clusters(clusters, offsets):
    clusters = [list(item) for item in clusters]
    mentions = [item for sublist in clusters for item in sublist]
    result = None
    for mention in mentions:
        if offsets[0] >= mention[0] and offsets[1] <= mention[1]:
            if result == None:
                result = mention
            else:
                if result[1] - result[0] > mention[1] - mention[0]:
                    result = mention
    return result


def evaluate_coref_scores(pred, gold, mention_to_pred, mention_to_gold):
    evaluator = OfficialCoNLL2012CorefEvaluator()

    for p, g, m2p, m2g in zip(pred, gold, mention_to_pred, mention_to_gold):
        evaluator.update(p, g, m2p, m2g)
    result = []
    for metric in ["muc", "b_cubed", "ceafe", "conll2012"]:
        result.append(dict(zip(["precision", "recall", "f1_score"], evaluator.get_prf(metric))))
    return result


def model_predictions_with_dataloader(model, test_dataloader, device):
    model.to(device)
    model.eval()
    predictions = []

    for batch in tqdm(test_dataloader, desc="Test", total=test_dataloader.__len__()):
        output = model.model(
            stage="test",
            input_ids=batch["input_ids"].to(device),
            attention_mask=batch["attention_mask"].to(device),
            eos_mask=batch["eos_mask"].to(device),
            gold_starts=batch["gold_starts"].to(device),
            gold_mentions=batch["gold_mentions"].to(device),
            gold_clusters=batch["gold_clusters"].to(device),
            tokens=batch["tokens"],
            subtoken_map=batch["subtoken_map"],
            new_token_map=batch["new_token_map"],
            singletons=False,
        )

        clusters_predicted = original_token_offsets(
            clusters=output["pred_dict"]["clusters"],
            subtoken_map=batch["subtoken_map"][0],
            new_token_map=batch["new_token_map"][0],
        )
        predictions.append(clusters_predicted)
    return predictions


@hydra.main(config_path="../conf", config_name="root")
def main(conf: omegaconf.DictConfig):
    evaluate(conf)


if __name__ == "__main__":
    main()



import torch
import numpy as np

from transformers import AutoTokenizer

from maverick.models.pl_modules import BasePLModule
from maverick.common.util import *
from maverick.common.constants import *
from maverick.models import *

from transformers.utils.hub import cached_file as hf_cached_file


class Maverick:
    # put the pip package online
    def __init__(self, hf_name_or_path="sapienzanlp/maverick-mes-ontonotes", device="cuda"):
        self.device = device
        path = self.__get_model_path__(hf_name_or_path)
        self.model = BasePLModule.load_from_checkpoint(path, _recursive_=False, map_location=self.device)
        self.model = self.model.eval()
        self.model = self.model.model
        self.tokenizer = self.__get_model_tokenizer__()

    def __get_model_path__(self, hf_name_or_path):
        try:
            print(hf_name_or_path, "loading")
            path = hf_cached_file(hf_name_or_path, "weights.ckpt")
        except:
            print(hf_name_or_path, "not found on huggingface, loading from local path ")
            path = hf_name_or_path
        return path

    def __get_model_tokenizer__(self):
        tokenizer = AutoTokenizer.from_pretrained(self.model.encoder_hf_model_name, use_fast=True, add_prefix_space=True)
        special_tokens_dict = {"additional_special_tokens": ["[SPEAKER_START]", "[SPEAKER_END]"]}
        tokenizer.add_special_tokens(special_tokens_dict)
        return tokenizer

    def __sample_type__(self, sample):
        if isinstance(sample, str):
            result = "text"
        if isinstance(sample, list):
            result = "word_tokenized"
            if len(sample) != 0 and isinstance(sample[0], list):
                result = "sentence_tokenized"
        return result

    def preprocess(self, sample, speakers=None):
        type = self.__sample_type__(sample)
        char_offsets = None
        if type == "text":
            nlp = download_load_spacy()
            char_offsets = []
            sentences = []
            off = 0
            s = sent_tokenize(sample)
            for sent, sentence in zip(nlp.pipe(s), s):
                char_offsets.append([(off + tok.idx, off + tok.idx + len(tok.text) - 1) for tok in sent])
                sentences.append([tok.text for tok in sent])
                off += len(sentence) + 1
            char_offsets = flatten(char_offsets)
            tokens = flatten(sentences)
            eos_len = [len(value) for value in sentences]
            eos = [sum(eos_len[0 : (i[0] + 1)]) for i in enumerate(eos_len)]
        elif type == "word_tokenized":
            nlp = download_load_spacy()
            tokens = sample
            eos = [idx + 1 for idx, tok in enumerate(tokens) if tok == "."]
            if len(eos) == 0 or eos[-1] != len(tokens):
                eos.append(len(tokens))
        elif type == "sentence_tokenized":
            sentences = sample
            tokens = flatten(sentences)
            eos_len = [len(value) for value in sentences]
            eos = [sum(eos_len[0 : (i[0] + 1)]) for i in enumerate(eos_len)]
        if speakers == None:
            speakers = ["-"] * len(tokens)
        else:
            speakers = flatten(speakers)
        return tokens, eos, speakers, char_offsets

    # takes length of sequence (int) and eos_indices ([])
    # returns len x len zeros matrix with 1 in pos (start, all possible ends)
    def eos_mask(self, input_ids_len, eos_indices):
        mask = np.zeros((input_ids_len, input_ids_len))
        prec = 0
        for eos_idx in eos_indices:
            for i in range(prec, eos_idx + 2):
                for j in range(prec, eos_idx + 2):
                    mask[i][j] = 1
            prec = eos_idx
        mask = np.triu(mask)
        return mask

    # enable speakers usage (only when we have tokenized input)
    @torch.no_grad()
    def predict(self, sample, singletons=False, add_gold_clusters=None, predefined_mentions=None, speakers=None):
        tokens, eos_indices, speakers, char_offsets = self.preprocess(sample, speakers)  # [[w1,w2,w3...], []]
        tokenized = self.tokenize(tokens, eos_indices, speakers, predefined_mentions, add_gold_clusters)

        output = self.model(
            stage="test",
            input_ids=torch.tensor(tokenized["input_ids"]).unsqueeze(0).to(self.device),
            attention_mask=torch.tensor(tokenized["attention_mask"]).unsqueeze(0).to(self.device),
            eos_mask=torch.tensor(tokenized["eos_mask"]).unsqueeze(0).to(self.device),
            tokens=[tokenized["tokens"]],
            subtoken_map=[tokenized["subtoken_map"]],
            new_token_map=[tokenized["new_token_map"]],
            singletons=singletons,
            add=tokenized["added"],
            gold_mentions=(
                None
                if tokenized["gold_mentions"] == None
                else torch.tensor(
                    self.create_mention_matrix(
                        len(tokenized["input_ids"]),
                        tokenized["gold_mentions"],
                    )
                )
                .unsqueeze(0)
                .to(self.device)
            ),
        )

        clusters_predicted = original_token_offsets(
            clusters=output["pred_dict"]["clusters"],
            subtoken_map=tokenized["subtoken_map"],
            new_token_map=tokenized["new_token_map"],
        )
        result = {}
        result["tokens"] = tokens
        result["clusters_token_offsets"] = clusters_predicted
        result["clusters_char_offsets"] = None
        result["clusters_token_text"] = [
            [" ".join(tokens[span[0] : span[1] + 1]) for span in cluster] for cluster in clusters_predicted
        ]
        result["clusters_char_text"] = None
        if char_offsets != None:
            result["clusters_char_offsets"] = [
                [(char_offsets[span[0]][0], char_offsets[span[1]][1]) for span in cluster] for cluster in clusters_predicted
            ]

            # result["clusters_char_text"] = [
            #     [sample[char_offsets[span[0]][0] : char_offsets[span[1]][1] + 1] for span in cluster]
            #     for cluster in clusters_predicted
            # ]

        return result

    def create_mention_matrix(self, input_ids_len, mentions):
        matrix = np.zeros((input_ids_len, input_ids_len))
        for start_bpe_idx, end_bpe_idx in mentions:
            matrix[start_bpe_idx][end_bpe_idx] = 1
        return matrix

    def tokenize(self, tokens, eos_indices, speakers=None, gold_mentions=None, add_gold_clusters=None):
        token_to_new_token_map = []  # len() = len(tokens), contains indices of original sequence to new sequence
        new_token_map = []  # len() = len(new_tokens), contains indices of new sequence
        new_tokens = []  # contains new tokens
        last_speaker = None

        for idx, (token, speaker) in enumerate(zip(tokens, speakers)):
            if last_speaker != speaker:
                new_tokens += ["[SPEAKER_START]", speaker, "[SPEAKER_END]"]
                new_token_map += [None, None, None]
                last_speaker = speaker
            token_to_new_token_map.append(len(new_tokens))
            new_token_map.append(idx)
            new_tokens.append(token)

        encoded_text = self.tokenizer(new_tokens, add_special_tokens=True, is_split_into_words=True)
        if gold_mentions != None:
            gold_mentions = [
                (
                    encoded_text.word_to_tokens(token_to_new_token_map[start]).start,
                    encoded_text.word_to_tokens(token_to_new_token_map[end]).end - 1,
                )
                for start, end in gold_mentions
            ]

        addedd = None
        if add_gold_clusters != None:
            addedd = [
                [
                    (
                        encoded_text.word_to_tokens(token_to_new_token_map[start]).start,
                        encoded_text.word_to_tokens(token_to_new_token_map[end]).end - 1,
                    )
                    for start, end in cluster
                ]
                for cluster in add_gold_clusters
            ]

        eos_indices = [
            encoded_text.word_to_tokens(token_to_new_token_map[eos - 1]).start
            for eos in eos_indices
            if encoded_text.word_to_tokens(token_to_new_token_map[eos - 1]) != None
        ]
        output = {
            "tokens": tokens,
            "input_ids": encoded_text["input_ids"],
            "attention_mask": encoded_text["attention_mask"],
            "subtoken_map": encoded_text.word_ids(),
            "new_token_map": new_token_map,
            "eos_mask": self.eos_mask(len(encoded_text["input_ids"]), eos_indices),
            "gold_mentions": gold_mentions,
            "added": addedd,
        }
        return output



from functools import partial
from typing import Any, Union, List, Optional, Sequence

import hydra
import pytorch_lightning as pl
import torch
from omegaconf import DictConfig
from pytorch_lightning.utilities.types import EVAL_DATALOADERS
from torch.utils.data import DataLoader, Dataset


class BasePLDataModule(pl.LightningDataModule):
    def __init__(
        self,
        dataset: DictConfig,
        batch_sizes: DictConfig,
        num_workers: DictConfig,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.dataset = dataset
        self.num_workers = num_workers
        self.batch_sizes = batch_sizes

        # data
        self.train_dataset: Optional[Dataset] = None
        self.val_dataset: Optional[Dataset] = None
        self.test_dataset: Optional[Dataset] = None

    def setup(self, stage: Optional[str] = None):
        if stage == "fit" or stage is None:
            self.val_dataset = hydra.utils.instantiate(self.dataset.val)
            self.train_dataset = hydra.utils.instantiate(self.dataset.train)
        if stage == "test" or stage is None:
            self.test_dataset = hydra.utils.instantiate(self.dataset.test)

    def train_dataloader(self, *args, **kwargs) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            shuffle=True,
            batch_size=self.batch_sizes.train,
            num_workers=self.num_workers.train,
            collate_fn=self.train_dataset.collate_fn,
        )

    def val_dataloader(self, *args, **kwargs) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            shuffle=False,
            batch_size=self.batch_sizes.val,
            num_workers=self.num_workers.val,
            collate_fn=self.val_dataset.collate_fn,
        )

    def test_dataloader(self, *args, **kwargs) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            shuffle=False,
            batch_size=self.batch_sizes.test,
            num_workers=self.num_workers.test,
            collate_fn=self.test_dataset.collate_fn,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(" f"{self.dataset=}, " f"{self.num_workers=}, " f"{self.batch_sizes=})"



from maverick.models.maverick_model import Maverick
from maverick.models.model_mes import Maverick_mes
from maverick.models.model_s2e import Maverick_s2e
from maverick.models.model_incr import Maverick_incr



import argparse
import json
import os
import tempfile
import zipfile
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from maverick.utils.loggingl import get_console_logger

logger = get_console_logger()

import huggingface_hub


def get_md5(path: Path):
    """
    Get the MD5 value of a path.
    """
    import hashlib

    with path.open("rb") as fin:
        data = fin.read()
    return hashlib.md5(data).hexdigest()


def create_info_file(tmpdir: Path):
    # logger.debug("Computing md5 of model.zip")
    md5 = get_md5(tmpdir / "model.zip")
    date = datetime.now()

    # logger.debug("Dumping info.json file")
    with (tmpdir / "info.json").open("w") as f:
        json.dump(dict(md5=md5, upload_date=date), f, indent=2)


def zip_run(
    dir_path: Union[str, os.PathLike],
    tmpdir: Union[str, os.PathLike],
    zip_name: str = "model.zip",
) -> Path:
    # logger.debug(f"zipping {dir_path} to {tmpdir}")
    # creates a zip version of the provided dir_path
    run_dir = Path(dir_path)
    zip_path = tmpdir / zip_name

    with zipfile.ZipFile(zip_path, "w") as zip_file:
        # fully zip the run directory maintaining its structure
        for file in run_dir.rglob("*.*"):
            if file.is_dir():
                continue

            zip_file.write(file, arcname=file.relative_to(run_dir))

    return zip_path


def get_logged_in_username():
    token = huggingface_hub.HfFolder.get_token()
    if token is None:
        raise ValueError("No HuggingFace token found. You need to execute `huggingface-cli login` first!")
    api = huggingface_hub.HfApi()
    user = api.whoami(token=token)
    return user["name"]


def upload(
    model_dir: Union[str, os.PathLike],
    model_name: str,
    filenames: Optional[list[str]] = None,
    organization: str | None = None,
    repo_name: str | None = None,
    commit: str | None = None,
    archive: bool = False,
):
    token = huggingface_hub.HfFolder.get_token()
    if token is None:
        raise ValueError("No HuggingFace token found. You need to execute `huggingface-cli login` first!")

    repo_id = repo_name or model_name
    if organization is not None:
        repo_id = f"{organization}/{repo_id}"
    with tempfile.TemporaryDirectory() as tmpdir:
        api = huggingface_hub.HfApi()
        repo_url = api.create_repo(
            token=token,
            repo_id=repo_id,
            exist_ok=True,
        )
        repo = huggingface_hub.Repository(str(tmpdir), clone_from=repo_url, use_auth_token=token)

        tmp_path = Path(tmpdir)
        if archive:
            # otherwise we zip the model_dir
            # logger.debug(f"Zipping {model_dir} to {tmp_path}")
            zip_run(model_dir, tmp_path)
            create_info_file(tmp_path)
        else:
            # if the user wants to upload a transformers model, we don't need to zip it
            # we just need to copy the files to the tmpdir
            # logger.debug(f"Copying {model_dir} to {tmpdir}")
            # copy only the files that are needed
            if filenames is not None:
                for filename in filenames:
                    os.system(f"cp {model_dir}/{filename} {tmpdir}")
            else:
                os.system(f"cp -r {model_dir}/* {tmpdir}")

        # this method automatically puts large files (>10MB) into git lfs
        repo.push_to_hub(commit_message=commit or "initial commit")


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("model_dir", help="The directory of the model you want to upload")
    parser.add_argument("model_name", help="The model you want to upload")
    parser.add_argument(
        "--organization",
        help="the name of the organization where you want to upload the model",
    )
    parser.add_argument(
        "--repo_name",
        help="Optional name to use when uploading to the HuggingFace repository",
    )
    parser.add_argument("--commit", help="Commit message to use when pushing to the HuggingFace Hub")
    parser.add_argument(
        "--archive",
        action="store_true",
        help="""
            Whether to compress the model directory before uploading it.
            If True, the model directory will be zipped and the zip file will be uploaded.
            If False, the model directory will be uploaded as is.""",
    )
    return parser.parse_args()


upload(**vars(parse_args()))



from maverick import Maverick

a = Maverick()
# r = a.predict("Hi i am Giuliano, i like pasta. Giuliano, my cool name")
print(
    a.predict(
        [
            [
                "--",
                "basically",
                ",",
                "it",
                "was",
                "unanimously",
                "agreed",
                "upon",
                "by",
                "the",
                "various",
                "relevant",
                "parties",
                ".",
            ],
            [
                "To",
                "express",
                "its",
                "determination",
                ",",
                "the",
                "Chinese",
                "securities",
                "regulatory",
                "department",
                "compares",
                "this",
                "stock",
                "reform",
                "to",
                "a",
                "die",
                "that",
                "has",
                "been",
                "cast",
                ".",
            ],
            [
                "It",
                "takes",
                "time",
                "to",
                "prove",
                "whether",
                "the",
                "stock",
                "reform",
                "can",
                "really",
                "meet",
                "expectations",
                ",",
                "and",
                "whether",
                "any",
                "deviations",
                "that",
                "arise",
                "during",
                "the",
                "stock",
                "reform",
                "can",
                "be",
                "promptly",
                "corrected",
                ".",
            ],
            ["Dear", "viewers", ",", "the", "China", "News", "program", "will", "end", "here", "."],
            ["This", "is", "Xu", "Li", "."],
            ["Thank", "you", "everyone", "for", "watching", "."],
            ["Coming", "up", "is", "the", "Focus", "Today", "program", "hosted", "by", "Wang", "Shilin", "."],
            ["Good-bye", ",", "dear", "viewers", "."],
        ],
        # predefined_mentions=[[16, 16], [19, 23], [42, 44], [57, 59], [25, 27], [83, 83], [82, 82]],
        add_gold_clusters=[[[1, 1], [2, 2]]],
        singletons=False,
    )
)



import numpy as np
import hydra.utils
import torch

from typing import Tuple
from typing import Dict, Union
from torch.utils.data import Dataset
from transformers import AutoTokenizer
from torch.utils.data import Dataset
from datasets import load_from_disk
from datasets import Dataset as dt

import maverick.common.util as util


NULL_ID_FOR_COREF = -1


class OntonotesDataset(Dataset):
    def __init__(self, name: str, path: str, batch_size, processed_dataset_path, tokenizer, **kwargs):
        super().__init__()
        self.stage = name
        self.path = path
        self.batch_size = batch_size
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer, use_fast=True, add_prefix_space=True)
        self.max_doc_len = kwargs.get("max_doc_len", None)
        special_tokens_dict = {"additional_special_tokens": ["[SPEAKER_START]", "[SPEAKER_END]"]}
        self.tokenizer.add_special_tokens(special_tokens_dict)
        try:
            self.set = load_from_disk(hydra.utils.get_original_cwd() + "/" + processed_dataset_path + "/")
        except:
            self.set = dt.from_pandas(util.ontonotes_to_dataframe(path))
            if self.stage == "train":
                if "clusters" not in self.set.column_names:
                    print("Training set has to include cluster information")
                # self.set = self.set.map(self.cut_document_to_length, batched=False)
                self.set = self.prepare_data(self.set)
            self.set = self.set.map(self.encode, batched=False)
            self.set = self.set.remove_columns(column_names=["speakers"])
            if self.stage != "test":
                self.set.save_to_disk(hydra.utils.get_original_cwd() + "/" + processed_dataset_path + "/")

    def prepare_data(self, set):
        return set.filter(lambda x: len(self.tokenizer(x["tokens"])["input_ids"]) <= self.max_doc_len)

    def cut_document_to_length(self, set_element):
        encoded_text = self.tokenizer(set_element["tokens"], add_special_tokens=True, is_split_into_words=True)
        if len(encoded_text["input_ids"]) <= self.max_doc_len + 1:
            result = set_element
        else:
            last_index_input_id_in_sentence = encoded_text.token_to_word(self.max_doc_len)
            eos_indices = [end for end in set_element["EOS_indices"] if end < last_index_input_id_in_sentence]
            last_sentence_end = eos_indices[-3]
            result = {
                "doc_key": set_element["doc_key"],
                "tokens": set_element["tokens"][:last_sentence_end],
                "speakers": set_element["speakers"][:last_sentence_end],
                "EOS_indices": eos_indices[:-3],
            }
            new_clusters = []
            for cluster in set_element["clusters"]:
                new_cluster = []
                for span in cluster:
                    if span[1] < last_sentence_end:
                        new_cluster.append(span)
                if len(new_cluster) >= 1:
                    new_clusters.append(new_cluster)
            result["clusters"] = new_clusters
        return result

    # missing genre information
    def _tokenize(self, tokens, clusters, speakers, eos_indices):
        token_to_new_token_map = []  # len() = len(tokens), contains indices of original sequence to new sequence
        new_token_map = []  # len() = len(new_tokens), contains indices of new sequence
        new_tokens = []  # contains new tokens
        last_speaker = None

        for idx, (token, speaker) in enumerate(zip(tokens, speakers)):
            if last_speaker != speaker:
                new_tokens += ["[SPEAKER_START]", speaker, "[SPEAKER_END]"]
                new_token_map += [None, None, None]
                last_speaker = speaker
            token_to_new_token_map.append(len(new_tokens))
            new_token_map.append(idx)
            new_tokens.append(token)

        for cluster in clusters:
            for start, end in cluster:
                assert tokens[start : end + 1] == new_tokens[token_to_new_token_map[start] : token_to_new_token_map[end] + 1]

        encoded_text = self.tokenizer(new_tokens, add_special_tokens=True, is_split_into_words=True)
        clusters = [
            [
                (
                    encoded_text.word_to_tokens(token_to_new_token_map[start]).start,
                    encoded_text.word_to_tokens(token_to_new_token_map[end]).end - 1,
                )
                for start, end in cluster
            ]
            for cluster in clusters
        ]
        eos_indices = [
            encoded_text.word_to_tokens(token_to_new_token_map[eos - 1]).start
            for eos in eos_indices
            if encoded_text.word_to_tokens(token_to_new_token_map[eos - 1]) != None
        ]
        output = {
            "tokens": tokens,
            "input_ids": encoded_text["input_ids"],
            "attention_mask": encoded_text["attention_mask"],
            "gold_clusters": clusters,
            "subtoken_map": encoded_text.word_ids(),
            "new_token_map": new_token_map,
            "EOS_indices": eos_indices,
        }
        return output

    def encode(self, example):
        if "clusters" not in example:
            example["clusters"] = []
        encoded = self._tokenize(
            example["tokens"],
            example["clusters"],
            example["speakers"],
            example["EOS_indices"],
        )  # debug when no clusters
        encoded["num_clusters"] = len(encoded["gold_clusters"]) if encoded["gold_clusters"] else 0
        encoded["max_cluster_size"] = max(len(c) for c in encoded["gold_clusters"]) if encoded["gold_clusters"] else 0
        encoded["length"] = len(encoded["input_ids"])
        return encoded

    def __len__(self) -> int:
        return self.set.shape[0]

    def __getitem__(self, index) -> Union[Dict[str, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        return self.set[index]

    # takes length of sequence (int) and eos_indices ([])
    # returns len x len zeros matrix with 1 in pos (start, all possible ends)
    def eos_mask(self, input_ids_len, eos_indices):
        mask = np.zeros((input_ids_len, input_ids_len))
        prec = 0
        for eos_idx in eos_indices:
            for i in range(prec, eos_idx + 2):
                for j in range(prec, eos_idx + 2):
                    mask[i][j] = 1
            prec = eos_idx
        mask = np.triu(mask)
        return mask

    # takes length of sequence (int) and coreferences ([[()]])
    # returns len x len zeros matrix with 1 in pos (start, end)
    def create_mention_matrix(self, input_ids_len, coreferences):
        matrix = np.zeros((input_ids_len, input_ids_len))
        for cluster in coreferences:
            for start_bpe_idx, end_bpe_idx in cluster:
                matrix[start_bpe_idx][end_bpe_idx] = 1
        return matrix

    # takes length of sequence (int) and coreferences ([[()]])
    # returns len zeros matrix with 1 in start position
    def create_start_matrix(self, input_ids_len, coreferences):
        matrix = np.zeros((input_ids_len))
        for cluster in coreferences:
            for start_bpe_idx, end_bpe_idx in cluster:
                matrix[start_bpe_idx] = 1
        return matrix

    # pad don't pad the rest, and is slow, think about something else
    def collate_fn(self, batch):
        batch = self.tokenizer.pad(batch)
        output = {
            "input_ids": torch.tensor(batch["input_ids"]),
            "attention_mask": torch.tensor(batch["attention_mask"]),
            "eos_mask": torch.tensor(self.eos_mask(len(batch["input_ids"][0]), batch["EOS_indices"][0])).unsqueeze(0),
            "gold_mentions": torch.tensor(
                self.create_mention_matrix(len(batch["input_ids"][0]), batch["gold_clusters"][0])
            ).unsqueeze(0),
            "gold_starts": torch.tensor(
                self.create_start_matrix(len(batch["input_ids"][0]), batch["gold_clusters"][0])
            ).unsqueeze(0),
        }

        max_num_clusters, max_max_cluster_size = max(batch["num_clusters"]), max(batch["max_cluster_size"])
        if max_num_clusters == 0:
            padded_clusters = []
        else:
            padded_clusters = [
                pad_clusters(cluster, max_num_clusters, max_max_cluster_size) for cluster in batch["gold_clusters"]
            ]

        output["gold_clusters"] = torch.tensor(padded_clusters)
        output["tokens"] = batch["tokens"]
        output["doc_key"] = batch["doc_key"]
        output["subtoken_map"] = batch["subtoken_map"]
        output["new_token_map"] = batch["new_token_map"]
        output["eos_indices"] = torch.tensor(batch["EOS_indices"])
        output["singletons"] = "litbank" in self.path or "preco" in self.path
        return output


def pad_clusters_inside(clusters, max_cluster_size):
    return [cluster + [(NULL_ID_FOR_COREF, NULL_ID_FOR_COREF)] * (max_cluster_size - len(cluster)) for cluster in clusters]


def pad_clusters_outside(clusters, max_num_clusters):
    return clusters + [[]] * (max_num_clusters - len(clusters))


def pad_clusters(clusters, max_num_clusters, max_cluster_size):
    clusters = pad_clusters_outside(clusters, max_num_clusters)
    clusters = pad_clusters_inside(clusters, max_cluster_size)
    return clusters



import pandas as pd
import spacy
import hydra
import torch
import pytorch_lightning as pl

from nltk import sent_tokenize
from torch.nn import Module, Linear, LayerNorm, Dropout, GELU

from maverick.common.constants import *


def get_category_id(mention, antecedent):
    mention, mention_pronoun_id = mention
    antecedent, antecedent_pronoun_id = antecedent

    if mention_pronoun_id > -1 and antecedent_pronoun_id > -1:
        if mention_pronoun_id == antecedent_pronoun_id:
            return CATEGORIES["pron-pron-comp"]
        else:
            return CATEGORIES["pron-pron-no-comp"]

    if mention_pronoun_id > -1 or antecedent_pronoun_id > -1:
        return CATEGORIES["pron-ent"]

    if mention == antecedent:
        return CATEGORIES["match"]

    union = mention.union(antecedent)
    if len(union) == max(len(mention), len(antecedent)):
        return CATEGORIES["contain"]

    return CATEGORIES["other"]


def get_pronoun_id(span):
    if len(span) == 1:
        span = list(span)
        if span[0] in PRONOUNS_GROUPS:
            return PRONOUNS_GROUPS[span[0]]
    return -1


def flatten(l):
    return [item for sublist in l for item in sublist]


def ontonotes_to_dataframe(file_path):
    # read file
    df = pd.read_json(hydra.utils.get_original_cwd() + "/" + file_path, lines=True)
    # ontonotes is split into words and sentences, jin sentences
    if "sentences" in df.columns:
        df["tokens"] = df["sentences"].apply(lambda x: flatten(x))
    elif "text" in df.columns:  # te
        nlp = spacy.load("en_core_web_sm", exclude=["tagger", "parser", "lemmatizer", "ner", "textcat"])
        texts = df["text"].tolist()
        df["sentences"] = [[[tok.text for tok in s] for s in nlp.pipe(s)] for s in [sent_tokenize(text) for text in texts]]

        nlp = spacy.load("en_core_web_sm", exclude=["tagger", "parser", "lemmatizer", "ner", "textcat"])

        df["tokens"] = [[tok.text for tok in doc] for doc in nlp.pipe(texts)]
    # compute end of sequence indices
    if "preco" in file_path:
        df["EOS_indices"] = df["tokens"].apply(lambda x: [i + 1 for i, token in enumerate(x) if token == "."])
    else:
        df["EOS_lengths"] = df["sentences"].apply(lambda x: [len(value) for value in x])
        df["EOS_indices"] = df["EOS_lengths"].apply(lambda x: [sum(x[0 : (i[0] + 1)]) for i in enumerate(x)])
    # add speakers
    if "speakers" in df.columns and "wkc" not in file_path and "q" not in file_path:
        df["speakers"] = df["speakers"].apply(lambda x: flatten(x))
    else:
        df["speakers"] = df["tokens"].apply(lambda x: ["-"] * len(x))

    if "clusters" in df.columns:
        df = df[["doc_key", "tokens", "speakers", "clusters", "EOS_indices"]]
    else:
        df = df[["doc_key", "tokens", "speakers", "EOS_indices"]]
    df = df.dropna()
    df = df.reset_index(drop=True)
    return df


def extract_mentions_to_clusters(gold_clusters):
    mention_to_gold = {}
    for gc in gold_clusters:
        for mention in gc:
            mention_to_gold[mention] = gc
    return mention_to_gold


def original_token_offsets(clusters, subtoken_map, new_token_map):
    return [
        tuple(
            [
                (
                    new_token_map[subtoken_map[start]],
                    new_token_map[subtoken_map[end]],
                )
                for start, end in cluster
                if subtoken_map[start] is not None
                and subtoken_map[end] is not None  # only happens first evals, model predicts <s> as mentions
            ]
        )
        for cluster in clusters
    ]


def unpad_gold_clusters(gold_clusters):
    new_gold_clusters = []
    for batch in gold_clusters:
        new_gold_clusters = []
        for cluster in batch:
            new_cluster = []
            for span in cluster:
                if span[0].item() != -1:
                    new_cluster.append((span[0].item(), span[1].item()))
            if len(new_cluster) != 0:
                new_gold_clusters.append(tuple(new_cluster))
    return new_gold_clusters


class FullyConnectedLayer(Module):
    def __init__(self, input_dim, output_dim, hidden_size, dropout_prob):
        super(FullyConnectedLayer, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.dropout_prob = dropout_prob

        self.dense1 = Linear(self.input_dim, hidden_size)
        self.dense = Linear(hidden_size, self.output_dim)
        self.layer_norm = LayerNorm(hidden_size)
        self.activation_func = GELU()
        self.dropout = Dropout(self.dropout_prob)

    def forward(self, inputs):
        temp = inputs
        temp = self.dense1(temp)
        temp = self.dropout(temp)
        temp = self.activation_func(temp)
        temp = self.layer_norm(temp)
        temp = self.dense(temp)
        return temp


class RepresentationLayer(torch.nn.Module):
    def __init__(self, type, input_dim, output_dim, hidden_dim, **kwargs) -> None:
        super(RepresentationLayer, self).__init__()
        self.hidden_dim = hidden_dim
        self.lt = type
        if type == "Linear":
            self.layer = Linear(input_dim, output_dim)
        elif type == "FC":
            self.layer = FullyConnectedLayer(input_dim, output_dim, hidden_dim, dropout_prob=0.2)
        elif type == "LSTM-left":
            self.layer = torch.nn.LSTM(input_size=input_dim, hidden_size=output_dim, bidirectional=True)
        elif type == "LSTM-right":
            self.layer = torch.nn.LSTM(input_size=input_dim, hidden_size=output_dim, bidirectional=True)
        elif type == "LSTM-bidirectional":
            self.layer = torch.nn.LSTM(input_size=input_dim, hidden_size=output_dim / 2, bidirectional=True)
        elif type == "Conv1d":
            self.layer = torch.nn.Conv1d(input_size=input_dim, hidden_size=output_dim, kernel_size=7, stride=1, padding=3)
            self.dropout = Dropout(0.2)

    def forward(self, inputs):
        if self.lt == "Linear":
            return self.layer(inputs)
        elif self.lt == "FC":
            return self.layer(inputs)
        elif self.lt == "LSTM-left":
            return self.layer(inputs)[0][: self.hidden_dim]
        elif self.lt == "LSTM-right":
            return self.layer(inputs)[0][self.hidden_dim :]
        elif self.lt == "LSTM-bidirectional":
            return self.layer(inputs)[0]
        elif self.lt == "Conv1d":
            return self.layer(self.dropout(inputs))


def download_load_spacy():
    try:
        nlp = spacy.load("en_core_web_sm", exclude=["tagger", "parser", "lemmatizer", "ner", "textcat"])
    except:
        from spacy.cli import download
        import nltk

        nltk.download("punkt")
        download("en_core_web_sm")
        nlp = spacy.load("en_core_web_sm", exclude=["tagger", "parser", "lemmatizer", "ner", "textcat"])
    return nlp



from typing import Any

import hydra
import pytorch_lightning as pl
import torch

from torchmetrics import *
import transformers
from transformers import Adafactor

from maverick.common.metrics import *
from maverick.common.util import *
from maverick.models.model_incr import *
from maverick.models.model_mes import Maverick_mes
from maverick.models.model_s2e import *


class BasePLModule(pl.LightningModule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.save_hyperparameters()
        try:
            self.model = hydra.utils.instantiate(self.hparams.model)
        except:
            self.hparams.model["_target_"] = "maverick." + self.hparams.model["_target_"]
            self.model = hydra.utils.instantiate(self.hparams.model)
        self.train_step_predictions = []
        self.train_step_gold = []
        self.validation_step_predictions = []
        self.validation_step_gold = []
        self.test_step_predictions = []
        self.test_step_gold = []

    def forward(self, batch) -> dict:
        output_dict = self.model(batch, "forward")
        return output_dict

    def evaluate(self, predictions, golds):
        mention_evaluator = OfficialMentionEvaluator()
        start_evaluator = OfficialMentionEvaluator()
        cluster_mention_evaluator = OfficialMentionEvaluator()
        coref_evaluator = OfficialCoNLL2012CorefEvaluator()
        result = {}

        for pred, gold in zip(predictions, golds):
            if "start_idxs" in pred.keys():
                starts_pred = pred["start_idxs"][0].tolist()
                starts_gold = (gold["gold_starts"][0] == 1).nonzero(as_tuple=False).squeeze(-1).tolist()
                start_evaluator.update(starts_pred, starts_gold)

            if "mention_idxs" in pred.keys():
                mentions_pred = [tuple(p) for p in pred["mention_idxs"][0].tolist()]
                mentions_gold = [tuple(g) for g in (gold["gold_mentions"][0] == 1).nonzero(as_tuple=False).tolist()]
                mention_evaluator.update(mentions_pred, mentions_gold)

            if "clusters" in pred.keys():
                pred_clusters = pred["clusters"]
                gold_clusters = gold["gold_clusters"]
                mention_to_gold_clusters = extract_mentions_to_clusters(gold_clusters)
                mention_to_predicted_clusters = extract_mentions_to_clusters(pred_clusters)

                coref_evaluator.update(
                    pred_clusters,
                    gold_clusters,
                    mention_to_predicted_clusters,
                    mention_to_gold_clusters,
                )

                cluster_mention_evaluator.update(
                    [item for sublist in gold_clusters for item in sublist],
                    [item for sublist in pred_clusters for item in sublist],
                )
        p, r, f1 = start_evaluator.get_prf()
        result.update(
            {
                "start_f1_score": f1,
                "start_precision": p,
                "start_recall": r,
            }
        )
        p, r, f1 = mention_evaluator.get_prf()
        result.update({"mention_f1_score": f1, "mention_precision": p, "mention_recall": r})
        p, r, f1 = cluster_mention_evaluator.get_prf()
        result.update(
            {
                "cluster_mention_f1_score": f1,
                "cluster_mention_precision": p,
                "cluster_mention_recall": r,
            }
        )

        for metric in ["muc", "b_cubed", "ceafe", "conll2012"]:
            p, r, f1 = coref_evaluator.get_prf(metric)
            result.update(
                {
                    metric + "_f1_score": f1,
                    metric + "_precision": p,
                    metric + "_recall": r,
                }
            )
        return result

    def training_step(self, batch: dict, batch_idx: int) -> torch.Tensor:
        output = self.model(
            stage="train",
            input_ids=batch["input_ids"],
            attention_mask=batch["attention_mask"],
            eos_mask=batch["eos_mask"],
            gold_starts=batch["gold_starts"],
            gold_mentions=batch["gold_mentions"],
            gold_clusters=batch["gold_clusters"],
            tokens=batch["tokens"],
            subtoken_map=batch["subtoken_map"],
            new_token_map=batch["new_token_map"],
            singletons=batch["singletons"],
        )
        self.log_dict({"train/" + k: v for k, v in output["loss_dict"].items()}, on_step=True)
        return output["loss"]

    def validation_step(self, batch: dict, batch_idx: int):
        output = self.model(
            stage="val",
            input_ids=batch["input_ids"],
            attention_mask=batch["attention_mask"],
            eos_mask=batch["eos_mask"],
            gold_starts=batch["gold_starts"],
            gold_mentions=batch["gold_mentions"],
            gold_clusters=batch["gold_clusters"],
            tokens=batch["tokens"],
            subtoken_map=batch["subtoken_map"],
            new_token_map=batch["new_token_map"],
            singletons=batch["singletons"],
        )
        self.log_dict({"val/" + k: v for k, v in output["loss_dict"].items()})
        output["pred_dict"]["clusters"] = original_token_offsets(
            clusters=output["pred_dict"]["clusters"],
            subtoken_map=batch["subtoken_map"][0],
            new_token_map=batch["new_token_map"][0],
        )
        self.validation_step_predictions.append(output["pred_dict"])
        self.validation_step_gold.append(
            {
                "gold_starts": batch["gold_starts"].cpu(),
                "gold_mentions": batch["gold_mentions"].cpu(),
                "gold_clusters": original_token_offsets(
                    clusters=unpad_gold_clusters(batch["gold_clusters"].cpu()),
                    subtoken_map=batch["subtoken_map"][0],
                    new_token_map=batch["new_token_map"][0],
                ),
            }
        )

    def on_validation_epoch_end(self):
        self.log_dict(
            {"val/" + k: v for k, v in self.evaluate(self.validation_step_predictions, self.validation_step_gold).items()}
        )
        self.validation_step_predictions = []
        self.validation_step_gold = []

    def test_step(self, batch: dict, batch_idx: int) -> Any:
        output = self.model(
            stage="test",
            input_ids=batch["input_ids"],
            attention_mask=batch["attention_mask"],
            eos_mask=batch["eos_mask"],
            eos_indices=batch["eos_indices"],
            tokens=batch["tokens"],
            subtoken_map=batch["subtoken_map"],
            new_token_map=batch["new_token_map"],
            singletons=batch["singletons"],
        )
        self.log_dict({"test/" + k: v for k, v in output["loss_dict"].items()})
        output["pred_dict"]["clusters"] = original_token_offsets(
            clusters=output["pred_dict"]["clusters"],
            subtoken_map=batch["subtoken_map"][0],
            new_token_map=batch["new_token_map"][0],
        )
        self.test_step_predictions.append(output["pred_dict"])
        self.test_step_gold.append(
            {
                "gold_starts": batch["gold_starts"].cpu(),
                "gold_mentions": batch["gold_mentions"].cpu(),
                "gold_clusters": original_token_offsets(
                    clusters=unpad_gold_clusters(batch["gold_clusters"].cpu()),
                    subtoken_map=batch["subtoken_map"][0],
                    new_token_map=batch["new_token_map"][0],
                ),
            }
        )

    def on_test_epoch_end(self):
        self.log_dict({"test/" + k: v for k, v in self.evaluate(self.test_step_predictions, self.test_step_gold).items()})
        self.test_step_predictions = []
        self.test_step_gold = []

    def configure_optimizers(self):
        if self.hparams.opt == "RAdam":
            opt = hydra.utils.instantiate(self.hparams.RAdam, params=self.parameters())
            return opt
        else:
            return self.custom_opt()

    def custom_opt(self):
        no_decay = ["bias", "LayerNorm.weight"]
        head_params = ["representaion", "classifier"]

        model_decay = [
            p
            for n, p in self.model.named_parameters()
            if not any(hp in n for hp in head_params) and not any(nd in n for nd in no_decay)
        ]
        model_no_decay = [
            p
            for n, p in self.model.named_parameters()
            if not any(hp in n for hp in head_params) and any(nd in n for nd in no_decay)
        ]
        head_decay = [
            p
            for n, p in self.model.named_parameters()
            if any(hp in n for hp in head_params) and not any(nd in n for nd in no_decay)
        ]
        head_no_decay = [
            p for n, p in self.model.named_parameters() if any(hp in n for hp in head_params) and any(nd in n for nd in no_decay)
        ]

        head_learning_rate = 3e-4
        lr = 2e-5
        wd = 0.01
        optimizer_grouped_parameters = [
            {"params": model_decay, "lr": lr, "weight_decay": wd},
            {"params": model_no_decay, "lr": lr, "weight_decay": 0.0},
            {"params": head_decay, "lr": head_learning_rate, "weight_decay": wd},
            {"params": head_no_decay, "lr": head_learning_rate, "weight_decay": 0.0},
        ]
        optimizer = Adafactor(optimizer_grouped_parameters, scale_parameter=False, relative_step=False)
        scheduler = transformers.get_linear_schedule_with_warmup(
            optimizer,
            num_warmup_steps=self.hparams.lr_scheduler.num_training_steps * 0.1,
            num_training_steps=self.hparams.lr_scheduler.num_training_steps,
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1,
            },
        }



PRONOUNS_GROUPS = {
    "i": 0,
    "me": 0,
    "my": 0,
    "mine": 0,
    "myself": 0,
    "you": 1,
    "your": 1,
    "yours": 1,
    "yourself": 1,
    "yourselves": 1,
    "he": 2,
    "him": 2,
    "his": 2,
    "himself": 2,
    "she": 3,
    "her": 3,
    "hers": 3,
    "herself": 3,
    "it": 4,
    "its": 4,
    "itself": 4,
    "we": 5,
    "us": 5,
    "our": 5,
    "ours": 5,
    "ourselves": 5,
    "they": 6,
    "them": 6,
    "their": 6,
    "themselves": 6,
    "that": 7,
    "this": 7,
}

STOPWORDS = {
    "'s",
    "a",
    "all",
    "an",
    "and",
    "at",
    "for",
    "from",
    "in",
    "into",
    "more",
    "of",
    "on",
    "or",
    "some",
    "the",
    "these",
    "those",
}

MODELS = ["mes", "s2e", "incr", None]
DATASETS = ["ontonotes", "litbank", "preco"]
CATEGORIES = {"pron-pron-comp": 0, "pron-pron-no-comp": 1, "pron-ent": 2, "match": 3, "contain": 4, "other": 5}



from setuptools import setup

setup(
    name="maverick-coref",
    version="0.9.6",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/sapienzanlp/maverick-coref",
    packages=["maverick", "maverick.models", "maverick.data", "maverick.utils", "maverick.common"],
    install_requires=[
        "pytorch-lightning>=1.8.0",
        "pandas>=1.3.5",
        "hydra-core>=1.2",
        "transformers>=4.34",
        "spacy>=3.7.2",
        "jsonlines==4.0.0",
        "sentencepiece==0.1.99",
        "protobuf==3.20",
        "nltk==3.8.1",
        "scipy>=1.10",
        "huggingface-hub>=0.19.0",
    ],
    python_requires=">=3.8.0",
    author="Giuliano Martinelli",
    author_email="giuliano.martinelli97@gmail.com",
    zip_safe=False,
)



import numpy as np
from collections import Counter
from scipy.optimize import linear_sum_assignment


# Mention evaluation performed as scores on set of mentions.
class OfficialMentionEvaluator:
    def __init__(self):
        self.tp, self.fp, self.fn = 0, 0, 0

    # takes predicted mentions and gold mentions:
    # a mention is  [[(wstart, wend)]]
    # calculates tp, fp, and fn as set operations
    # returns prf using formula
    def update(self, predicted_mentions, gold_mentions):
        predicted_mentions = set(predicted_mentions)
        gold_mentions = set(gold_mentions)

        self.tp += len(predicted_mentions & gold_mentions)
        self.fp += len(predicted_mentions - gold_mentions)
        self.fn += len(gold_mentions - predicted_mentions)

    def get_f1(self):
        pr = self.get_precision()
        rec = self.get_recall()
        return 2 * pr * rec / (pr + rec) if pr + rec > 0 else 0.0

    def get_recall(self):
        return self.tp / (self.tp + self.fn) if (self.tp + self.fn) > 0 else 0.0

    def get_precision(self):
        return self.tp / (self.tp + self.fp) if (self.tp + self.fp) > 0 else 0.0

    def get_prf(self):
        return self.get_precision(), self.get_recall(), self.get_f1()


# CoNLL-2012 metrics https://aclanthology.org/W12-4501.pdf#page=19
class OfficialCoNLL2012CorefEvaluator(object):
    def __init__(self):
        self.evaluators = [OfficialEvaluator(m) for m in (muc, b_cubed, ceafe)]
        self.metrics = {"muc": 0, "b_cubed": 1, "ceafe": 2}

    # pred & golds : [[((wstart, wend),()), (())...]],  mention_to_pred and mention_to_gold: [{(wstart,wend):((wstart, wend)...)}]
    def update(self, pred, gold, mention_to_predicted, mention_to_gold):
        for e in self.evaluators:
            e.update(pred, gold, mention_to_predicted, mention_to_gold)

    def get_f1(self, metric):
        if metric == "conll2012":
            return sum(e.get_f1() for e in self.evaluators) / len(self.evaluators)
        else:
            return self.evaluators[self.metrics[metric]].get_f1()

    def get_recall(self, metric):
        if metric == "conll2012":
            return sum(e.get_recall() for e in self.evaluators) / len(self.evaluators)
        else:
            return self.evaluators[self.metrics[metric]].get_recall()

    def get_precision(self, metric):
        if metric == "conll2012":
            return sum(e.get_precision() for e in self.evaluators) / len(self.evaluators)
        else:
            return self.evaluators[self.metrics[metric]].get_precision()

    def get_prf(self, metric):
        return self.get_precision(metric), self.get_recall(metric), self.get_f1(metric)


class OfficialEvaluator(object):
    def __init__(self, metric, beta=1):
        self.p_num = 0
        self.p_den = 0
        self.r_num = 0
        self.r_den = 0
        self.metric = metric
        self.beta = beta

    def update(self, predicted, gold, mention_to_predicted, mention_to_gold):
        if self.metric == ceafe:
            pn, pd, rn, rd = self.metric(predicted, gold)
        else:
            pn, pd = self.metric(predicted, mention_to_gold)
            rn, rd = self.metric(gold, mention_to_predicted)
        self.p_num += pn
        self.p_den += pd
        self.r_num += rn
        self.r_den += rd

    def get_f1(self):
        return f1(self.p_num, self.p_den, self.r_num, self.r_den, beta=self.beta)

    def get_recall(self):
        return 0 if self.r_num == 0 else self.r_num / float(self.r_den)

    def get_precision(self):
        return 0 if self.p_num == 0 else self.p_num / float(self.p_den)

    def get_prf(self):
        return self.get_precision(), self.get_recall(), self.get_f1()

    def get_counts(self):
        return self.p_num, self.p_den, self.r_num, self.r_den


def f1(p_num, p_den, r_num, r_den, beta=1):
    p = 0 if p_den == 0 else p_num / float(p_den)
    r = 0 if r_den == 0 else r_num / float(r_den)
    return 0 if p + r == 0 else (1 + beta * beta) * p * r / (beta * beta * p + r)


# mention based metric (Bagga and Baldwin, 1998)
def b_cubed(clusters, mention_to_gold):
    num, dem = 0, 0

    for c in clusters:
        gold_counts = Counter()
        correct = 0
        for m in c:
            if m in mention_to_gold:
                gold_counts[tuple(mention_to_gold[m])] += 1
        for c2, count in gold_counts.items():
            correct += count * count
        if len(c) != 0:
            num += correct / float(len(c))
            dem += len(c)

    return num, dem


# link based metric (Vilain et al. 2995).
def muc(clusters, mention_to_gold):
    tp, p = 0, 0
    for c in clusters:
        p += len(c) - 1
        tp += len(c)
        linked = set()
        for m in c:
            if m in mention_to_gold:
                linked.add(mention_to_gold[m])
            else:
                tp -= 1
        tp -= len(linked)
    return tp, p


# Entity based metric (Luo, 2005)
def ceafe(clusters, gold_clusters):
    # clusters = [c for c in clusters if len(c) != 1]
    scores = np.zeros((len(gold_clusters), len(clusters)))
    for i in range(len(gold_clusters)):
        for j in range(len(clusters)):
            scores[i, j] = phi4(gold_clusters[i], clusters[j])
    row_ind, col_ind = linear_sum_assignment(-scores)
    similarity = sum(scores[row_ind, col_ind])
    return similarity, len(clusters), similarity, len(gold_clusters)


def phi4(c1, c2):
    return 2 * len([m for m in c1 if m in c2]) / float(len(c1) + len(c2))


def lea(clusters, mention_to_gold):
    num, dem = 0, 0

    for c in clusters:
        if len(c) == 1:
            continue

        common_links = 0
        all_links = len(c) * (len(c) - 1) / 2.0
        for i, m in enumerate(c):
            if m in mention_to_gold:
                for m2 in c[i + 1 :]:
                    if m2 in mention_to_gold and mention_to_gold[m] == mention_to_gold[m2]:
                        common_links += 1

        num += len(c) * common_links / float(all_links)
        dem += len(c)

    return num, dem



import logging
import sys
import threading
from typing import Optional

from rich.console import Console

_lock = threading.Lock()
_default_handler: Optional[logging.Handler] = None

_default_log_level = logging.WARNING

# fancy logger
_console = Console()


def _get_library_name() -> str:
    return __name__.split(".")[0]


def _get_library_root_logger() -> logging.Logger:
    return logging.getLogger(_get_library_name())


def _configure_library_root_logger() -> None:

    global _default_handler

    with _lock:
        if _default_handler:
            # This library has already configured the library root logger.
            return
        _default_handler = logging.StreamHandler()  # Set sys.stderr as stream.
        _default_handler.flush = sys.stderr.flush

        # Apply our default configuration to the library root logger.
        library_root_logger = _get_library_root_logger()
        library_root_logger.addHandler(_default_handler)
        library_root_logger.setLevel(_default_log_level)
        library_root_logger.propagate = False


def _reset_library_root_logger() -> None:

    global _default_handler

    with _lock:
        if not _default_handler:
            return

        library_root_logger = _get_library_root_logger()
        library_root_logger.removeHandler(_default_handler)
        library_root_logger.setLevel(logging.NOTSET)
        _default_handler = None


def set_log_level(level: int, logger: logging.Logger = None) -> None:
    """
    Set the log level.
    Args:
        level (:obj:`int`):
            Logging level.
        logger (:obj:`logging.Logger`):
            Logger to set the log level.
    """
    if not logger:
        _configure_library_root_logger()
        logger = _get_library_root_logger()
    logger.setLevel(level)


def get_logger(
    name: Optional[str] = None,
    level: Optional[int] = None,
    formatter: Optional[str] = None,
) -> logging.Logger:
    """
    Return a logger with the specified name.
    """

    if name is None:
        name = _get_library_name()

    _configure_library_root_logger()

    if level is not None:
        set_log_level(level)

    if formatter is None:
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    _default_handler.setFormatter(formatter)

    return logging.getLogger(name)


def get_console_logger():
    return _console
